"""Blob.

References to files stored in a on-disk repository
"""
import hashlib
import uuid
from pathlib import Path
from typing import Optional

import sqlalchemy as sa
from sqlalchemy.event import listens_for
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.unitofwork import UOWTransaction
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer

from abilian.core.models.base import Model
from abilian.core.sqlalchemy import UUID, JSONDict


class Blob(Model):
    """Model for storing large file content.

    Files are stored on-disk, named after their uuid. Repository is
    located in instance folder/data/files.
    """

    __tablename__ = "blob"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    uuid = Column(UUID(), unique=True, nullable=False, default=uuid.uuid4)
    meta = Column(JSONDict(), nullable=False, default=dict)

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.uuid is None:
            self.uuid = uuid.uuid4()

        if self.meta is None:
            self.meta = {}

        if value is not None:
            self.value = value

    @property
    def file(self) -> Optional[Path]:
        """Return :class:`pathlib.Path` object used for storing value."""
        from abilian.services.repository import session_repository as repository

        return repository.get(self, self.uuid)

    @property
    def size(self) -> int:
        """Return size in bytes of value."""
        f = self.file
        return f.stat().st_size if f is not None else 0

    @property
    def value(self) -> bytes:
        """Binary value content."""
        v = self.file
        return v.open("rb").read() if v is not None else v

    @value.setter
    def value(self, value: bytes):
        """Store binary content to applications's repository and update
        `self.meta['md5']`.

        :param:content: bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        from abilian.services.repository import session_repository as repository

        repository.set(self, self.uuid, value)
        self.meta["md5"] = str(hashlib.md5(self.value).hexdigest())

        if hasattr(value, "filename"):
            filename = value.filename
            if isinstance(filename, bytes):
                filename = filename.decode("utf-8")
            self.meta["filename"] = filename

        if hasattr(value, "content_type"):
            self.meta["mimetype"] = value.content_type

    @value.deleter
    def value(self) -> None:
        """Remove value from repository."""
        from abilian.services.repository import session_repository as repository

        repository.delete(self, self.uuid)

    @property
    def md5(self) -> str:
        """Return md5 from meta, or compute it if absent."""
        md5 = self.meta.get("md5")
        if md5 is None:
            md5 = str(hashlib.md5(self.value).hexdigest())

        return md5

    def __bool__(self) -> bool:
        """A blob is considered falsy if it has no file."""
        return self.file is not None and self.file.exists()

    # Py3k compat
    __nonzero__ = __bool__


@listens_for(sa.orm.Session, "after_flush")
def _blob_propagate_delete_content(
    session: Session, flush_context: UOWTransaction
) -> None:
    deleted = (obj for obj in session.deleted if isinstance(obj, Blob))
    for blob in deleted:
        del blob.value



""""""
from datetime import datetime, timedelta

from flask import Flask
from pytest import raises

from abilian.core.entities import Entity
from abilian.core.models.comment import Comment, is_commentable, register
from abilian.core.sqlalchemy import SQLAlchemy


@register
class CommentableContent(Entity):
    pass


def test_commentable_interface() -> None:
    assert is_commentable(CommentableContent)

    instance = CommentableContent(name="test instance")
    assert not is_commentable(instance)  # not in DB: no id

    instance.id = 42
    assert is_commentable(instance)
    assert not is_commentable(object)
    assert not is_commentable(object())


def test_cannot_register_non_entities() -> None:
    class Dummy:
        pass

    with raises(ValueError):
        register(Dummy)


def test_default_ordering(app: Flask, db: SQLAlchemy) -> None:
    commentable = CommentableContent(name="commentable objet")
    db.session.add(commentable)

    now = datetime.now()
    c1 = Comment(entity=commentable, body="comment #1")
    c1.created_at = now - timedelta(10)
    db.session.flush()
    c2 = Comment(entity=commentable, body="comment #2")
    c2.created_at = now - timedelta(5)
    db.session.flush()

    query = Comment.query.filter(Comment.entity == commentable)
    assert query.all() == [c1, c2]

    c1.created_at = c2.created_at
    c2.created_at = c1.created_at - timedelta(5)
    assert query.all() == [c2, c1]



import os
import sys
import tempfile
from pathlib import Path
from typing import Iterator, Union
from warnings import warn

from magic import Magic
from pytest import fixture, mark

from abilian.services.conversion.handlers import HAS_LIBREOFFICE, HAS_PDFTOTEXT
from abilian.services.conversion.service import Converter

mime_sniffer = Magic(mime=True)
encoding_sniffer = Magic(mime_encoding=True)

# FIXME: tests that rely on OOo are disabled until we fix stability issues.


@fixture
def converter() -> Iterator[Union[Iterator, Iterator[Converter]]]:
    from abilian.services.conversion import converter as c

    cache_dir = Path(tempfile.mkdtemp(suffix="unittest"))
    tmp_dir = Path(tempfile.mkdtemp(suffix="unittest"))
    c.init_work_dirs(cache_dir, tmp_dir)
    yield c

    c.clear()


def read_file(fn: str, mode: str = "rb") -> bytes:
    return (Path(__file__).parent / "dummy_files" / fn).open(mode).read()


# To text
@mark.skipif(not HAS_PDFTOTEXT, reason="requires poppler or poppler-util")
def test_pdf_to_text(converter):
    blob = read_file("onepage.pdf")
    text = converter.to_text("", blob, "application/pdf")
    assert text


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_word_to_text(converter: Converter) -> None:
    blob = read_file("test.doc")
    text = converter.to_text("", blob, "application/msword")
    assert text


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_wordx_to_text(converter: Converter) -> None:
    blob = read_file("test.docx")
    text = converter.to_text("", blob, "application/msword")
    assert text


# def test_excel_to_text(converter):
#     blob = read_file("test.xls")
#     text = converter.to_text("", blob, "application/excel")


# To PDF
@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_odt_to_pdf(converter: Converter) -> None:
    blob = read_file("test.odt")
    pdf = converter.to_pdf("", blob, "application/vnd.oasis.opendocument.text")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


@mark.skipif(not HAS_LIBREOFFICE, reason="requires libreoffice")
def test_word_to_pdf(converter: Converter) -> None:
    blob = read_file("test.doc")
    pdf = converter.to_pdf("", blob, "application/msword")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


@mark.skipif(
    sys.platform.startswith("linux"), reason="maybe bug due to ImagMagik settings"
)
def test_image_to_pdf(converter: Converter) -> None:
    blob = read_file("picture.jpg")
    pdf = converter.to_pdf("", blob, "image/jpeg")
    assert "application/pdf" == mime_sniffer.from_buffer(pdf)


# To images
@mark.skipif(not HAS_PDFTOTEXT, reason="requires poppler or poppler-util")
def test_pdf_to_images(converter: Converter) -> None:
    if not os.popen("which pdftoppm").read().strip():
        warn("pdftoppm not found, skipping test")
        return
    blob = read_file("onepage.pdf")
    image = converter.to_image("", blob, "application/pdf", 0)
    assert "image/jpeg" == mime_sniffer.from_buffer(image)


@mark.skipif(
    not HAS_PDFTOTEXT or not HAS_LIBREOFFICE, reason="requires poppler or poppler-util"
)
def test_word_to_images(converter: Converter) -> None:
    blob = read_file("test.doc")
    image = converter.to_image("", blob, "application/msword", 0)
    assert "image/jpeg" == mime_sniffer.from_buffer(image)



import os

from flask import Flask

from abilian.core import signals


def init_hooks(app: Flask) -> None:
    @app.before_first_request
    def set_current_celery_app() -> None:
        """Listener for `before_first_request`.

        Set our celery app as current, so that task use the correct
        config. Without that tasks may use their default set app.
        """
        celery = app.extensions.get("celery")
        if celery:
            celery.set_current()

    @app.before_first_request
    def register_signals() -> None:
        signals.register_js_api.send(app)

    # def install_id_generator(sender, **kwargs):
    #     g.id_generator = count(start=1)
    #
    # appcontext_pushed.connect(install_id_generator)

    if os.environ.get("FLASK_VALIDATE_HTML"):
        # Workaround circular import
        from abilian.testing.validation import validate_response

        app.after_request(validate_response)



"""Modules that provide services.

They are implemented as Flask extensions (see:
http://flask.pocoo.org/docs/extensiondev/ )
"""
from flask import current_app

# This one must be imported first
from .base import Service, ServiceState

# Don't remove (used to force import order)
assert Service, ServiceState

from .activity import ActivityService
from .antivirus import service as antivirus
from .audit import audit_service
from .auth import AuthService
from .conversion import conversion_service, converter
from .indexing import service as index_service
from .preferences import preferences as preferences_service
from .repository import repository as repository_service
from .repository import session_repository as session_repository_service
from .security import security as security_service
from .settings import SettingsService
from .vocabularies import vocabularies as vocabularies_service

auth_service = AuthService()
activity_service = ActivityService()
settings_service = SettingsService()


def get_service(service: str) -> Service:
    return current_app.services[service]



""""""
import logging
import socket
from collections import OrderedDict, namedtuple
from pathlib import Path
from typing import Dict, Text

import redis
import sqlalchemy as sa
import sqlalchemy.dialects
import sqlalchemy.exc
from flask import (
    current_app,
    flash,
    redirect,
    render_template,
    request,
    session,
    url_for,
)

from abilian.core.commands import config as cmd_config
from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.core.util import unwrap
from abilian.services import get_service
from abilian.services.security import Admin, Anonymous
from abilian.web.blueprints import Blueprint

logger = logging.getLogger(__name__)
setup = Blueprint(
    "setup", __name__, allowed_roles=Anonymous, template_folder="templates"
)

# list supported dialects and detect unavailable ones due to missing dbapi
# module ('psycopg2' missing for example)

_dialects = OrderedDict((("sqlite", "SQLite (for demo)"), ("postgresql", "PostgreSQL")))

_dialects_unavailable: Dict[Text, Text] = OrderedDict()

for dialect, _label in _dialects.items():
    d = sa.dialects.registry.load(dialect)
    try:
        d.dbapi()
    except ImportError as e:
        _dialects_unavailable[dialect] = str(e)

# enumerate steps for left column progress
Step = namedtuple("SetupStep", ("name", "endpoint", "title", "description"))

_setup_steps = (
    Step("db", "step_db", "Setup Database", "Setup basic database connection"),
    Step("redis", "step_redis", "Setup Redis", "Redis connection"),
    Step(
        "site_info",
        "step_site_info",
        "Basic site informations",
        "Site name, admin email...",
    ),
    Step("admin_account", "step_admin_account", "Admin account", None),
    Step("finalize", "finalize", "Finalize", None),
)


@setup.before_request
def step_progress():
    session.permanent = False  # limit lifetime to browser session
    current_step = next_step = prev_step = None

    for step in _setup_steps:
        if current_step is not None:
            next_step = step
            break
        endpoint = f"{setup.name}.{step.endpoint}"
        if endpoint == request.endpoint:
            current_step = step
        else:
            prev_step = step

    request.setup_step_progress = {
        "current_step": current_step,
        "prev_step": prev_step,
        "next_step": next_step,
    }


@setup.context_processor
def common_context():
    ctx = {
        "setup_steps": _setup_steps,
        "dialects": _dialects,
        "dialects_unavailable": _dialects_unavailable,
        "validated_steps": session_get("validated", ()),
    }
    ctx.update(request.setup_step_progress)
    return ctx


# session helpers
def session_set(key, data):
    session.setdefault("abilian_setup", {})[key] = data


def session_get(key, default=None):
    return session.setdefault("abilian_setup", {}).get(key, default)


def session_clear():
    if "abilian_setup" in session:
        del session["abilian_setup"]


def step_validated(step_endpoint, valid=True):
    validated = set(session_get("validated", ()))
    if valid:
        validated.add(step_endpoint)
    else:
        if step_endpoint in validated:
            validated.remove(step_endpoint)

    session_set("validated", tuple(validated))


#
# DB Setup
#
@csrf.exempt
@setup.route("/", methods=["GET", "POST"])
def step_db():
    if request.method == "POST":
        return step_db_validate()
    return step_db_form()


def step_db_form():
    return render_template("setupwizard/step_db.html", data=session_get("db", {}))


def step_db_validate():
    form = request.form
    dialect = form["dialect"]
    username = form.get("username", "").strip()
    password = form.get("password", "").strip()
    host = form.get("host", "").strip()
    port = form.get("port", "").strip()
    database = form.get("database", "").strip()

    # build db_uri
    db_uri = f"{dialect}://"

    if dialect != "sqlite":
        if username:
            db_uri += username
            # check password only if we have a username
            if password:
                db_uri += f":{password}"

        # FIXME: it is an error to have a username and no host,
        # SA will interpret username:password as host:port
        if host:
            db_uri += f"@{host}"
            if port:
                db_uri += f":{port}"

    db_uri += "/"

    if dialect == "sqlite" and (not database or database == ":memory:"):
        database = str(Path(current_app.instance_path) / "data" / "sqlite.db")

    db_uri += database

    # store in session
    db_params = {
        "uri": db_uri,
        "dialect": dialect,
        "username": username,
        "password": password,
        "host": host,
        "port": port,
        "database": database,
    }
    session_set("db", db_params)

    # test connection
    engine = sa.create_engine(db_uri)
    error = None
    try:
        conn = engine.connect()
    except sa.exc.OperationalError as e:
        error = str(e)
    else:
        conn.close()
        del conn

    engine.dispose()

    # FIXME: at this point we have not tested the connection is valid, i.e, we
    # have CREATE, SELECT, UPDATE, DELETE rights on the database

    if error:
        flash(error, "error")
        return step_db_form()

    step_validated("db")
    next_step = request.setup_step_progress["next_step"]
    endpoint = f"{setup.name}.{next_step.endpoint}"
    return redirect(url_for(endpoint))


#
# Redis
#
@csrf.exempt
@setup.route("/redis", methods=["GET", "POST"])
def step_redis():
    if request.method == "POST":
        return step_redis_validate()
    return step_redis_form()


def step_redis_form():
    return render_template("setupwizard/step_redis.html", data=session_get("redis", {}))


def step_redis_validate():
    form = request.form
    data = {
        "host": form.get("host", "localhost").strip(),
        "port": form.get("port", "").strip() or "6379",
        "db": form.get("db", "").strip() or "1",
    }

    for k in ("port", "db"):
        try:
            data[k] = int(data[k])
        except ValueError:
            pass

    data["uri"] = "redis://{host}:{port}/{db}".format(**data)
    session_set("redis", data)
    error = None

    try:
        r = redis.StrictRedis(host=data["host"], port=data["port"], db=data["db"])
    except Exception as e:
        error = "Connection error, check parameters"
        raise e

    try:
        r.info()
    except redis.exceptions.InvalidResponse:
        error = (
            "Connection error: doesn't look like it's a redis server. "
            "Verify host and port are those of your redis server."
        )
    except redis.exceptions.ResponseError as e:
        error = f"Redis server response: {e}"
    except redis.exceptions.RedisError as e:
        error = f"Unknown redis error ({e})"

    if error:
        flash(error, "error")
        return step_redis_form()

    step_validated("redis")
    next_step = request.setup_step_progress["next_step"]
    endpoint = f"{setup.name}.{next_step.endpoint}"
    return redirect(url_for(endpoint))


#
# Site info
#
@csrf.exempt
@setup.route("/site_info", methods=["GET", "POST"])
def step_site_info():
    if request.method == "POST":
        return step_site_info_validate()
    return step_site_info_form()


def get_possible_hostnames():
    hostname = socket.gethostname()
    fqdn = socket.getfqdn()
    names = {"localhost": ["127.0.0.1"]}

    for name in (hostname, fqdn):
        try:
            name, aliases, ips = socket.gethostbyname_ex(name)
        except OSError:
            continue
        names.setdefault(name, []).extend(ips)
        for a in aliases:
            names.setdefault(a, []).extend(ips)

    return sorted(
        f"{name} ({', '.join(sorted(set(ips)))})" for name, ips in names.items()
    )


def step_site_info_form():
    cfg = current_app.config
    default_data = {
        "sitename": cfg.get("SITE_NAME", "") or "",
        "mailsender": cfg.get("MAIL_SENDER", "") or "",
    }
    return render_template(
        "setupwizard/step_site_info.html",
        data=session_get("site_info", default_data),
        suggested_hosts=get_possible_hostnames(),
    )


def step_site_info_validate():
    form = request.form
    data = {
        "sitename": form.get("sitename", "").strip(),
        "mailsender": form.get("mailsender", "").strip(),
        "server_mode": form.get("server_mode", "").strip(),
    }

    session_set("site_info", data)
    step_validated("site_info")
    next_step = request.setup_step_progress["next_step"]
    endpoint = f"{setup.name}.{next_step.endpoint}"
    return redirect(url_for(endpoint))


#
# Admin account creation
#
@csrf.exempt
@setup.route("/admin_account", methods=["GET", "POST"])
def step_admin_account():
    if request.method == "POST":
        return step_admin_account_validate()
    return step_admin_account_form()


def step_admin_account_form():
    return render_template(
        "setupwizard/step_admin_account.html", data=session_get("admin_account", {})
    )


def step_admin_account_validate():
    form = request.form
    password = form.get("password", "").strip()
    confirm_password = form.get("confirm_password", "").strip()
    admin_user = {
        "email": form.get("email", "").strip(),
        "name": form.get("name", "").strip(),
        "firstname": form.get("firstname", "").strip(),
        "password": password,
        "confirm_password": confirm_password,
    }
    session_set("admin_account", admin_user)

    if password != confirm_password:
        flash("Password fields don't match", "error")
        return step_admin_account_form()

    step_validated("admin_account")
    next_step = request.setup_step_progress["next_step"]
    endpoint = f"{setup.name}.{next_step.endpoint}"
    return redirect(url_for(endpoint))


#
# Finalize
#
@csrf.exempt
@setup.route("/finalize", methods=["GET", "POST"])
def finalize():
    validated = session_get("validated")
    assert all(s.name in validated for s in _setup_steps[:-1])

    if request.method == "POST":
        return finalize_validate()
    return finalize_form()


def finalize_form():
    file_location = Path(current_app.instance_path) / "config.py"
    return render_template("setupwizard/finalize.html", file_location=file_location)


def finalize_validate():
    config_file = Path(current_app.instance_path) / "config.py"
    logging_file = Path(current_app.instance_path) / "logging.yml"

    assert not config_file.exists()
    config = cmd_config.DefaultConfig(logging_file="logging.yml")
    config.SQLALCHEMY_DATABASE_URI = session_get("db")["uri"]

    redis_uri = session_get("redis")["uri"]
    config.REDIS_URI = redis_uri
    config.BROKER_URL = redis_uri
    config.CELERY_RESULT_BACKEND = redis_uri

    d = session_get("site_info")
    config.SITE_NAME = d["sitename"]
    config.MAIL_SENDER = d["mailsender"]

    is_production = d["server_mode"] == "production"
    config.PRODUCTION = is_production
    config.DEBUG = not is_production
    config.DEBUG_TB_ENABLED = config.DEBUG
    config.CELERY_ALWAYS_EAGER = not is_production

    cmd_config.write_config(config_file, config)
    cmd_config.maybe_write_logging(logging_file)

    admin_account = session_get("admin_account")
    # create a new app that will be configured with new config,
    # to create database and admin_user
    setup_app = unwrap(current_app)
    app = setup_app.__class__(
        setup_app.import_name,
        static_url_path=setup_app.static_url_path,
        static_folder=setup_app.static_folder,
        template_folder=setup_app.template_folder,
        instance_path=setup_app.instance_path,
    )
    with app.test_request_context("/setup/finalize"):
        app.create_db()
        db_session = db.session()
        admin = User(
            email=admin_account["email"],
            password=admin_account["password"],
            last_name=admin_account["name"],
            first_name=admin_account["firstname"],
            can_login=True,
        )
        db_session.add(admin)
        security = get_service("security")
        security.grant_role(admin, Admin)
        db_session.commit()

    session_clear()

    return render_template(
        "setupwizard/done.html", config_file=config_file, logging_file=logging_file
    )



""""""

from flask import Flask, current_app
from flask_login import current_user

from abilian.core.models import attachment as attachments
from abilian.services.security import READ, WRITE, security
from abilian.web import url_for

from .forms import AttachmentForm
from .views import UPLOAD_BUTTON
from .views import bp as blueprint

_MANAGER_ATTR = "__attachments_manager__"


class AttachmentExtension:
    """API for comments, installed as an application extension.

    It is also available in templates as `attachments`.
    """

    def __init__(self, app: Flask) -> None:
        app.extensions["attachments"] = self
        app.add_template_global(self, "attachments")
        app.register_blueprint(blueprint)

    def manager(self, obj):
        """Returns the :class:`AttachmentsManager` instance for this object."""
        manager = getattr(obj, _MANAGER_ATTR, None)
        if manager is None:
            manager = AttachmentsManager()
            setattr(obj.__class__, _MANAGER_ATTR, manager)

        return manager

    def supports_attachments(self, obj):
        return self.manager(obj).supports_attachments(obj)

    def for_entity(self, obj, check_support_attachments=False):
        return self.manager(obj).for_entity(
            obj, check_support_attachments=check_support_attachments
        )

    def has_attachments(self, obj):
        return self.manager(obj).has_attachments(obj)

    def count(self, obj):
        return self.manager(obj).count(obj)

    def get_form_context(self, obj):
        """Return a dict: form instance, action button, submit url...

        Used by macro m_attachment_form(entity)
        """
        return self.manager(obj).get_form_context(obj)


_DEFAULT_TEMPLATE = "macros/attachment_default.html"


class AttachmentsManager:
    """Allow customization of attachments form, display macros, etc.

    can be used as class decorator
    """

    Form = AttachmentForm
    macros_template = "macros/attachment.html"

    def __init__(
        self, Form=AttachmentForm, macros_template="macros/attachment_default.html"
    ):
        self.Form = Form
        self.macros_template = macros_template

    def __call__(self, Model):
        setattr(Model, _MANAGER_ATTR, self)
        return Model

    @property
    def macros(self):
        default_template = current_app.jinja_env.get_template(_DEFAULT_TEMPLATE)
        template = current_app.jinja_env.get_template(self.macros_template)
        default = default_template.module
        m = template.module
        return {
            "m_attachments": getattr(m, "m_attachments", default.m_attachments),
            "m_attachment": getattr(m, "m_attachment", default.m_attachment),
            "m_attachment_form": getattr(
                m, "m_attachment_form", default.m_attachment_form
            ),
        }

    def supports_attachments(self, obj):
        return attachments.supports_attachments(obj)

    def for_entity(self, obj, check_support_attachments=False):
        return attachments.for_entity(
            obj, check_support_attachments=check_support_attachments
        )

    def has_attachments(self, obj):
        return bool(self.for_entity(obj, check_support_attachments=True))

    def count(self, obj):
        return len(self.for_entity(obj, check_support_attachments=True))

    def get_form_context(self, obj):
        """Return a dict: form instance, action button, submit url...

        Used by macro m_attachment_form(entity)
        """
        return {
            "url": url_for("attachments.create", entity_id=obj.id),
            "form": self.Form(),
            "buttons": [UPLOAD_BUTTON],
        }

    #
    # current user capabilities
    #
    def can_view(self, entity):
        """True if user can view attachments on entity."""
        return security.has_permission(current_user, READ, obj=entity)

    def can_edit(self, entity):
        """True if user can edit attachments on entity."""
        return security.has_permission(current_user, WRITE, obj=entity)

    def can_create(self, entity):
        """True if user can add attachments."""
        return security.has_permission(current_user, WRITE, obj=entity)

    def can_delete(self, entity):
        """True if user can delete attachments."""
        return security.has_permission(current_user, WRITE, obj=entity)



from typing import Any


class PreferencePanel:
    """Base class for preference panels.

    Currently, this class does nothing. I may be useful in the future
    either as just a marker interface (for automatic plugin discovery /
    registration), or to add some common functionnalities. Otherwise, it
    will be removed.
    """

    id: str
    label: Any

    def is_accessible(self):
        return True

    def get(self):
        raise NotImplementedError

    def post(self):
        raise NotImplementedError



""""""
from datetime import datetime
from typing import Optional

import sqlalchemy as sa
import sqlalchemy.orm
from flask.blueprints import BlueprintSetupState
from flask_login import current_user
from werkzeug.exceptions import BadRequest

from abilian.core.entities import Entity
from abilian.core.models.comment import Comment, is_commentable
from abilian.core.util import utc_dt
from abilian.i18n import _, _l
from abilian.web import nav, url_for
from abilian.web.action import ButtonAction, actions
from abilian.web.blueprints import Blueprint
from abilian.web.views.object import (
    CANCEL_BUTTON,
    ObjectCreate,
    ObjectDelete,
    ObjectEdit,
)

from .forms import CommentForm

bp = Blueprint("comments", __name__, url_prefix="/comments")


def _default_comment_view(obj, obj_type, obj_id, **kwargs):
    entity = obj.entity
    return url_for(entity, _anchor=f"comment-{obj.id}")


@bp.record_once
def register_default_view(state: BlueprintSetupState) -> None:
    state.app.default_view.register(Comment, _default_comment_view)


COMMENT_BUTTON = ButtonAction("form", "edit", btn_class="primary", title=_l("Post"))


class BaseCommentView:
    Model = Comment
    Form = CommentForm

    #: commented entity
    entity: Optional[Entity] = None

    def init_object(self, args, kwargs):
        args, kwargs = super().init_object(args, kwargs)
        entity_id = kwargs.pop("entity_id", None)
        if entity_id is not None:
            self.entity = Entity.query.get(entity_id)

        if self.entity is None:
            raise BadRequest("No entity to comment")

        if not is_commentable(self.entity):
            raise BadRequest("This entity is not commentable")

        actions.context["object"] = self.entity
        return args, kwargs

    def view_url(self):
        kw = {}
        if self.obj and self.obj.id:
            kw["_anchor"] = f"comment-{self.obj.id}"
        return url_for(self.entity, **kw)

    def index_url(self):
        return self.view_url()

    @property
    def activity_target(self):
        return self.entity


class CommentEditView(BaseCommentView, ObjectEdit):

    _message_success = _l("Comment edited")

    def breadcrumb(self):
        label = _('Edit comment on "{title}"').format(title=self.entity.name)
        return nav.BreadcrumbItem(label=label)

    def get_form_buttons(self, *args, **kwargs):
        return [COMMENT_BUTTON, CANCEL_BUTTON]

    def after_populate_obj(self):
        obj_meta = self.obj.meta.setdefault("abilian.core.models.comment", {})
        history = obj_meta.setdefault("history", [])
        history.append(
            {
                "user_id": current_user.id,
                "user": str(current_user),
                "date": utc_dt(datetime.utcnow()).isoformat(),
            }
        )
        self.obj.meta.changed()


edit_view = CommentEditView.as_view("edit")
bp.route("/<int:entity_id>/<int:object_id>/edit")(edit_view)


class CommentCreateView(BaseCommentView, ObjectCreate):
    """"""

    _message_success = _l("Comment added")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_object(self, args, kwargs):
        args, kwargs = super().init_object(args, kwargs)
        self.obj.entity = self.entity
        session = sa.orm.object_session(self.entity)

        if session:
            sa.orm.session.make_transient(self.obj)

        return args, kwargs

    def breadcrumb(self):
        label = _('New comment on "{title}"').format(title=self.entity.name)
        return nav.BreadcrumbItem(label=label)

    def get_form_buttons(self, *args, **kwargs):
        return [COMMENT_BUTTON, CANCEL_BUTTON]


create_view = CommentCreateView.as_view("create")
bp.route("/<int:entity_id>/create")(create_view)


class CommentDeleteView(BaseCommentView, ObjectDelete):

    _message_success = _l("Comment deleted")


delete_view = CommentDeleteView.as_view("delete")
bp.route("/<int:entity_id>/<int:object_id>/delete")(delete_view)



"""Add a few specific filters to Jinja2."""
import re
from calendar import timegm
from datetime import date, datetime
from functools import wraps
from typing import Any, Callable, Optional, Union

import bleach
import dateutil.parser
import flask_babel as babel
from babel.dates import DateTimePattern, format_timedelta, parse_pattern
from deprecated import deprecated
from jinja2 import Environment, escape, evalcontextfilter
from jinja2.nodes import EvalContext
from markupsafe import Markup
from pytz import utc
from werkzeug.routing import BuildError

from abilian.core.util import local_dt, slugify, utc_dt

from .util import url_for


def autoescape(filter_func: Callable) -> Callable:
    """Decorator to autoescape result from filters."""

    @evalcontextfilter
    @wraps(filter_func)
    def _autoescape(
        eval_ctx: EvalContext, *args: str, **kwargs: Any
    ) -> Union[Markup, str]:
        result = filter_func(*args, **kwargs)
        if eval_ctx.autoescape:
            result = Markup(result)
        return result

    return _autoescape


@autoescape
def nl2br(value: str) -> Markup:
    """Replace newlines with <br />."""
    result = escape(value).replace("\n", Markup("<br />\n"))
    return result


_PARAGRAPH_RE = re.compile(r"(?:\r\n|\r|\n){2,}")


@autoescape
def paragraphs(value: str) -> str:
    """Blank lines delimitates paragraphs."""
    result = "\n\n".join(
        "<p>{}</p>".format(p.strip().replace("\n", Markup("<br />\n")))
        for p in _PARAGRAPH_RE.split(escape(value))
    )
    return result


def labelize(s: str) -> str:
    return " ".join([w.capitalize() for w in s.split("_")])


def filesize(d: Union[int, str]) -> Markup:
    if not isinstance(d, int):
        d = int(d)

    if d < 1000:
        s = f"{d:d}&nbsp;B"

    elif d < int(1e4):
        s = f"{d / 1000.0:.1f}&nbsp;kB"
    elif d < int(1e6):
        s = f"{d / 1000.0:.0f}&nbsp;kB"

    elif d < int(1e7):
        s = f"{d / 1000000.0:.1f}&nbsp;MB"
    elif d < int(1e9):
        s = f"{d / 1000000.0:.0f}&nbsp;MB"

    elif d < int(1e10):
        s = f"{d / 1000000000.0:.1f}&nbsp;GB"

    else:
        s = f"{d / 1000000000.0:.0f}&nbsp;GB"

    return Markup(s)


def roughsize(size: int, above: int = 20, mod: int = 10) -> str:
    """6 -> '6' 15 -> '15' 134 -> '130+'."""
    if size < above:
        return str(size)

    return f"{(size - size % mod):d}+"


def datetimeparse(s) -> Optional[datetime]:
    """Parse a string date time to a datetime object.

    Suitable for dates serialized with .isoformat()

    :return: None, or an aware datetime instance, tz=UTC.
    """
    try:
        dt = dateutil.parser.parse(s)
    except ValueError:
        return None

    return utc_dt(dt)


def age(
    dt: Optional[datetime],
    now: Optional[datetime] = None,
    add_direction: bool = True,
    date_threshold: Optional[Any] = None,
) -> str:
    """
    :param dt: :class:`datetime<datetime>` instance to format

    :param now: :class:`datetime<datetime>` instance to compare to `dt`

    :param add_direction: if `True`, will add "in" or "ago" (example for `en`
       locale) to time difference `dt - now`, i.e "in 9 min." or " 9min. ago"

    :param date_threshold: above threshold, will use a formated date instead of
       elapsed time indication. Supported values: "day".
    """
    # Fail silently for now XXX
    if not dt:
        return ""

    if not now:
        now = datetime.utcnow()

    locale = babel.get_locale()
    dt = utc_dt(dt)
    now = utc_dt(now)
    delta = dt - now

    if date_threshold is not None:
        dy, dw, dd = dt_cal = dt.isocalendar()
        ny, nw, nd = now_cal = now.isocalendar()

        if dt_cal != now_cal:
            # not same day
            remove_year = dy == ny
            date_fmt = locale.date_formats["long"].pattern
            time_fmt = locale.time_formats["short"].pattern
            fmt = locale.datetime_formats["medium"]

            if remove_year:
                date_fmt = date_fmt.replace("y", "").strip()
                # remove leading or trailing spaces, comma, etc...
                date_fmt = re.sub("^[^A-Za-z]*|[^A-Za-z]*$", "", date_fmt)

            fmt = fmt.format(time_fmt, date_fmt)
            return babel.format_datetime(dt, format=fmt)

    # don't use (flask.ext.)babel.format_timedelta: as of Flask-Babel 0.9 it
    # doesn't support "threshold" arg.
    return format_timedelta(
        delta,
        locale=locale,
        granularity="minute",
        threshold=0.9,
        add_direction=add_direction,
    )


def date_age(dt: Optional[datetime], now: Optional[datetime] = None) -> str:
    # Fail silently for now XXX
    if not dt:
        return ""

    formatted_date = babel.format_datetime(dt, format="yyyy-MM-dd HH:mm")
    return f"{formatted_date} ({age(dt, now)})"


@deprecated
def date_fmt(value, format="EE, d MMMM y"):
    """
    @deprecated: use flask_babel's dateformat filter instead.
    """
    if isinstance(value, date):
        return babel.format_date(value, format)
    else:
        return babel.format_date(local_dt(value), format)


def babel2datepicker(pattern: DateTimePattern) -> str:
    """Convert date format from babel (http://babel.pocoo.org/docs/dates/#date-
    fields)) to a format understood by bootstrap-datepicker."""
    if not isinstance(pattern, DateTimePattern):
        pattern = parse_pattern(pattern)

    map_fmt = {
        # days
        "d": "dd",
        "dd": "dd",
        "EEE": "D",
        "EEEE": "DD",
        "EEEEE": "D",  # narrow name => short name
        # months
        "M": "mm",
        "MM": "mm",
        "MMM": "M",
        "MMMM": "MM",
        # years
        "y": "yyyy",
        "yy": "yyyy",
        "yyy": "yyyy",
        "yyyy": "yyyy",
        # time picker format
        # hours
        "h": "%I",
        "hh": "%I",
        "H": "%H",
        "HH": "%H",
        # minutes,
        "m": "%M",
        "mm": "%M",
        # seconds
        "s": "%S",
        "ss": "%S",
        # am/pm
        "a": "%p",
    }

    return pattern.format % map_fmt


# Doesn't work yet. TZ issues.
def to_timestamp(dt):
    utc_datetime = dt.astimezone(utc)
    return timegm(utc_datetime.timetuple()) + utc_datetime.microsecond / 1e6


def abbrev(s: str, max_size: int) -> str:
    if len(s) <= max_size:
        return s
    else:
        h = max_size // 2 - 1
        return f"{s[0:h]}...{s[-h:]}"


def bool2check(val, true="\u2713", false=""):
    """Filter value as boolean and show check mark (✓) or nothing."""
    return true if val else false


@autoescape
def linkify(s: str) -> Markup:
    return Markup(bleach.linkify(s))


def obj_to_url(obj):
    """Find url for obj using :func:`url_for`, return empty string is not
    found.

    :func:`url_for` is also provided in jinja context, the filtering version is
    forgiving when `obj` has no default view set.
    """
    try:
        return url_for(obj)
    except BuildError:
        return ""


def init_filters(env: Environment) -> None:
    env.filters["nl2br"] = nl2br
    env.filters["paragraphs"] = paragraphs
    env.filters["date_age"] = date_age
    env.filters["datetimeparse"] = datetimeparse
    env.filters["age"] = age
    env.filters["date"] = date_fmt
    env.filters["babel2datepicker"] = babel2datepicker
    env.filters["to_timestamp"] = to_timestamp
    env.filters["url_for"] = obj_to_url
    env.filters["abbrev"] = abbrev
    env.filters["filesize"] = filesize
    env.filters["roughsize"] = roughsize
    env.filters["labelize"] = labelize
    env.filters["linkify"] = linkify
    env.filters["toslug"] = slugify
    env.filters["bool2check"] = bool2check



""""""
from typing import Any

import pytest
from pytest import raises

from .models import Setting, empty_value


def test_type_set() -> None:
    s = Setting()
    # registered base type: no failure
    s.type = "int"
    s.type = "bool"
    s.type = "json"
    s.type = "string"

    with raises(ValueError):
        s.type = "dummy type name"


OBJ = [1, 2, "été", {"1": "1", "2": "2"}]


@pytest.mark.parametrize(
    "type_,value",
    [
        ("int", 42),
        ("bool", True),
        ("bool", False),
        ("string", "test"),
        ("string", "bel été"),
        ("json", None),
        ("json", OBJ),
    ],
)
def test_set_get(type_: str, value: Any) -> None:
    s = Setting(key="key", type=type_)
    s.value = value
    assert s.value == value


def test_empty_value() -> None:
    s = Setting(key="key", type="json")
    assert s.value == empty_value


def test_service_facade(app, session):
    svc = app.services.get("settings")
    svc.set("key_1", 42, "int")
    session.flush()
    assert svc.get("key_1") == 42

    # new key with no type: raise error:
    with raises(ValueError):
        svc.set("key_err", 42)

    # key already with type_, this should not raise an error
    svc.set("key_1", 24)
    session.flush()
    assert svc.get("key_1") == 24

    svc.delete("key_1")
    session.flush()
    with raises(KeyError):
        svc.get("key_1")

    # delete: silent by default
    svc.delete("non_existent")

    # delete: non-silent
    with raises(KeyError):
        svc.delete("non_existent", silent=False)

    # tricky use case: ask key delete, set it later, then flush
    svc.set("key_1", 42, "int")
    session.flush()
    svc.delete("key_1")
    svc.set("key_1", 1)
    session.flush()
    assert svc.get("key_1") == 1

    # list keys
    svc.set("key_2", 2, "int")
    svc.set("other", "azerty", "string")
    session.flush()
    assert sorted(svc.keys()) == ["key_1", "key_2", "other"]
    assert sorted(svc.keys(prefix="key_")) == ["key_1", "key_2"]

    # as dict
    assert svc.as_dict() == {"other": "azerty", "key_1": 1, "key_2": 2}
    assert svc.as_dict(prefix="key_") == {"key_1": 1, "key_2": 2}


def test_namespace(app, session):
    svc = app.services.get("settings")
    ns = svc.namespace("test")
    ns.set("1", 42, "int")
    session.flush()
    assert ns.get("1") == 42
    assert svc.get("test:1") == 42

    ns.set("sub:2", 2, "int")
    svc.set("other", "not in NS", "string")
    session.flush()
    assert sorted(ns.keys()) == ["1", "sub:2"]
    assert sorted(svc.keys()) == ["other", "test:1", "test:sub:2"]

    # sub namespace test:sub:
    sub = ns.namespace("sub")
    assert sub.keys() == ["2"]
    assert sub.get("2") == 2

    sub.set("1", 1, "int")
    session.flush()
    assert sub.get("1") == 1
    assert ns.get("1") == 42
    assert sorted(svc.keys()) == ["other", "test:1", "test:sub:1", "test:sub:2"]

    # as dict
    assert sub.as_dict() == {"1": 1, "2": 2}
    assert ns.as_dict(prefix="sub:") == {"sub:1": 1, "sub:2": 2}
    assert ns.as_dict() == {"1": 42, "sub:1": 1, "sub:2": 2}
    assert svc.as_dict() == {
        "other": "not in NS",
        "test:1": 42,
        "test:sub:1": 1,
        "test:sub:2": 2,
    }

    # deletion
    sub.delete("1")
    sub.delete("2")
    session.flush()
    assert sub.keys() == []
    assert ns.keys() == ["1"]
    assert sorted(svc.keys()) == ["other", "test:1"]



from typing import Any, Callable, Collection, Dict, List, Optional, Union, cast

from flask_login import current_user
from wtforms.fields import Field

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.services import get_service
from abilian.services.security import (
    CREATE,
    READ,
    WRITE,
    Anonymous,
    Permission,
    Role,
    SecurityService,
)


class FormPermissions:
    """Form role/permission manager."""

    def __init__(
        self,
        default: Role = Anonymous,
        read: Union[None, Role, Collection[Role]] = None,
        write: Union[None, Role, Collection[Role]] = None,
        fields_read: Optional[Dict[str, Collection[Role]]] = None,
        fields_write: Optional[Dict[str, Collection[Role]]] = None,
        existing: Optional[Any] = None,
    ) -> None:
        """
        :param default: default roles when not specified for field. Can be:

            * a :class:`Role` or an iterable of :class:`Role`

            * a callable that returns a :class:`Role` or an iterable of
              :class:`Role`

            * a `dict` with :class:`Permission` instances for keys and one of other
              acceptable role spec.; a default entry `"default"` is required.

        :param read: global roles required for `READ` permission for whole form.

        :param write: global roles required for `WRITE` permission for whole form.
        """
        if isinstance(default, Role):
            default_dict = {"default": (default,)}
        elif isinstance(default, dict):
            if "default" not in default:
                raise ValueError('`default` parameter must have a "default" key')
            default_dict = default
        elif callable(default):
            default_dict = {"default": default}
        else:
            raise ValueError(
                "No valid value for `default`. Use a Role, an iterable "
                "of Roles, a callable, or a dict."
            )

        self.default = default_dict
        self.form = {}
        self.fields = {}

        if existing is not None:
            # copy existing formpermissions instance
            # maybe overwrite later with our definitions
            assert isinstance(existing, FormPermissions)
            for permission in (READ, WRITE):
                if permission in existing.form:
                    self.form[permission] = existing.form[permission]

            for field, mapping in existing.fields.items():
                f_map = self.fields[field] = {}
                for permission, roles in mapping.items():
                    f_map[permission] = roles

        for permission, roles in ((READ, read), (WRITE, write)):
            if roles is None:
                continue
            if isinstance(roles, Role):
                roles = (roles,)
            self.form[permission] = roles

        fields_defs = (
            (fields_read, READ),
            (fields_write, WRITE),
            (fields_write, CREATE),
        )  # checking against CREATE permission
        # at field level is the same as
        # WRITE permisssion
        for fields, permission in fields_defs:
            if fields:
                for field_name, allowed_roles in fields.items():
                    if isinstance(allowed_roles, Role):
                        allowed_roles = (allowed_roles,)
                    self.fields.setdefault(field_name, {})[permission] = allowed_roles

    def has_permission(
        self,
        permission: Permission,
        field: Optional[str] = None,
        obj: Union[None, Entity, object] = None,
        user: Optional[User] = None,
    ) -> bool:
        if user is None:
            user = cast(User, current_user)
        if obj is not None and not isinstance(obj, Entity):
            # permission/role can be set only on entities
            return True

        # FIXME: permission is not a str
        allowed_roles = (
            self.default[permission]
            if permission in self.default
            else self.default["default"]
        )
        definition = None

        def eval_roles(fun: Callable) -> List[Role]:
            return fun(permission=permission, field=field, obj=obj)

        if field is None:
            definition = self.form
        else:
            if isinstance(field, Field):
                field = field.name
            if field in self.fields:
                definition = self.fields[field]

        if definition and permission in definition:
            allowed_roles = definition[permission]

        if callable(allowed_roles):
            allowed_roles = eval_roles(allowed_roles)

        roles = []
        for r in allowed_roles:
            if callable(r):
                r = eval_roles(r)

            if isinstance(r, (Role, str)):
                roles.append(r)
            else:
                roles.extend(r)

        security: SecurityService = get_service("security")
        return security.has_role(user, role=roles, object=obj)



"""Extension for sending informations to upstream server."""
import typing
from typing import Any

from flask import Flask, _request_ctx_stack
from flask.signals import request_finished, request_started
from flask.wrappers import Response

from abilian.core.signals import user_loaded

if typing.TYPE_CHECKING:
    from abilian.core.models.subjects import User


class UpstreamInfo:
    """Extension for sending informations to upstream server."""

    def __init__(self, app: Flask = None) -> None:
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        app.extensions["upstream"] = self
        request_started.connect(self.request_started, sender=app)
        request_finished.connect(self.request_finished, sender=app)
        user_loaded.connect(self.user_loaded, sender=app)

        config = app.config
        config.setdefault("ABILIAN_UPSTREAM_INFO_ENABLED", False)
        for key in ("ABILIAN_UPSTREAM_INFO_DISCARD", "ABILIAN_UPSTREAM_INFO_INCLUDE"):
            val = config.get(key, ())
            if val is not None:
                val = frozenset(val)
            config[key] = val

    def request_started(self, app: Flask) -> None:
        _request_ctx_stack.top.upstream_info = {"Username": "Anonymous"}

    def request_finished(self, app: Flask, response: Response) -> None:
        info = _request_ctx_stack.top.upstream_info
        config = app.config
        default_enabled = bool(config["ABILIAN_UPSTREAM_INFO_ENABLED"])
        disabled = config["ABILIAN_UPSTREAM_INFO_DISCARD"]
        enabled = config["ABILIAN_UPSTREAM_INFO_INCLUDE"]

        for key, val in info.items():
            if (
                default_enabled
                and key in disabled
                or not default_enabled
                and key not in enabled
            ):
                continue

            header = f"X-{key}"
            response.headers[header] = val

    def user_loaded(self, app: Flask, user: "User", *args: Any, **kwargs: Any) -> None:
        _request_ctx_stack.top.upstream_info["Username"] = user.email


extension = UpstreamInfo()



""""""
import json
from functools import partial

from flask import request, url_for
from flask.ctx import AppContext
from flask.testing import FlaskClient
from sqlalchemy.orm import Session

from abilian.app import Application
from abilian.core.models.subjects import User
from abilian.services.auth import views


def test_get_redirect_target(app: Application, app_context: AppContext) -> None:
    get_redirect_target = views.get_redirect_target
    form_url = partial(url_for, "login.login_form")

    with app.test_request_context(form_url()):
        assert get_redirect_target() == ""
        url_root = request.url_root[:-1]

    with app.test_request_context(form_url(next="/")):
        assert get_redirect_target() == f"{url_root}/"

    # test "next" from referer
    referrer = f"{url_root}/some/path"
    with app.test_request_context(form_url(), headers=[("Referer", referrer)]):
        assert get_redirect_target() == referrer

    # don't cycle if coming from 'login.*' page, like this kind of cycle:
    # forgot password form ->  login page -> success
    # -> redirect(next = forgot password form) -> ...
    referrer = url_root + url_for("login.forgotten_pw")
    with app.test_request_context(form_url(), headers=[("Referer", referrer)]):
        assert get_redirect_target() == ""

    # test open redirect is forbidden
    with app.test_request_context(form_url(next="http://google.com/test")):
        assert get_redirect_target() == ""

    # open redirect through malicious construct and browser not checking
    # Location
    with app.test_request_context(form_url(next="/////google.com")):
        assert get_redirect_target() == f"{url_root}///google.com"


def test_login_post(session: Session, client: FlaskClient) -> None:
    kwargs = {"email": "User@domain.tld", "password": "azerty", "can_login": True}
    user = User(**kwargs)
    session.add(user)
    session.flush()

    response = client.post("/user/login", data=kwargs)
    assert response.status_code == 302

    # wrong password
    d = dict(kwargs)
    d["password"] = "wrong one"
    response = client.post("/user/login", data=d)
    assert response.status_code == 401

    # login disabled
    user.can_login = False
    session.flush()
    response = client.post("/user/login", data=kwargs)
    assert response.status_code == 401


def test_api_post(session: Session, client: FlaskClient) -> None:
    kwargs = {"email": "User@domain.tld", "password": "azerty", "can_login": True}
    user = User(**kwargs)
    session.add(user)
    session.flush()

    response = client.post(
        "/user/api/login", data=json.dumps(kwargs), content_type="application/json"
    )
    assert response.status_code == 200
    assert response.get_json() == {
        "email": "User@domain.tld",
        "username": "user@domain.tld",
        "fullname": "Unknown",
        "next_url": "",
    }

    response = client.post("/user/api/logout")
    assert response.status_code == 200


def test_forgotten_pw(app: Application, session: Session, client: FlaskClient) -> None:
    mail = app.extensions["mail"]
    kwargs = {"email": "User@domain.tld", "password": "azerty", "can_login": True}
    user = User(**kwargs)
    session.add(user)
    session.flush()

    payload = dict(**kwargs)
    del payload["password"]

    with mail.record_messages() as outbox:
        response = client.post("/user/forgotten_pw", data=payload)
        assert response.status_code == 302
        assert len(outbox) == 1

        msg = outbox[0]
        assert msg.subject == "Password reset instruction for Abilian Test"
        assert msg.recipients == ["User@domain.tld"]
        assert msg.cc == []



""""""
from typing import Any, Callable, Dict, Tuple

from flask import g, json, jsonify, redirect, render_template_string, request
from flask.views import MethodView as BaseView
from werkzeug.exceptions import HTTPException

from abilian.web.action import actions


class Redirect(HTTPException):
    pass


class View(BaseView):
    """Base class to use for all class based views.

    The view instance is accessible in :data:`g <flask.g>` and is set in
    :meth:`actions context <abilian.web.action.ActionRegistry.context>`.
    """

    @classmethod
    def as_view(cls, name: str, *class_args: Any, **class_kwargs: Any) -> Callable:
        return super().as_view(name, *class_args, **class_kwargs)

    def dispatch_request(self, *args: Any, **kwargs: Any) -> str:
        meth = getattr(self, request.method.lower(), None)
        # if the request method is HEAD and we don't have a handler for it
        # retry with GET
        if meth is None and request.method == "HEAD":
            meth = getattr(self, "get", None)
            assert meth is not None, f"Unimplemented method {request.method!r}"

        g.view = actions.context["view"] = self
        try:
            args, kwargs = self.prepare_args(args, kwargs)
            return meth(*args, **kwargs)
        except Redirect as exc:
            return exc.response

    def prepare_args(self, args, kwargs):
        """If view arguments need to be prepared it can be done here.

        A typical use case is to take an identifier, convert it to an
        object instance and maybe store it on view instance and/or
        replace identifier by object in arguments.
        """
        return args, kwargs

    def redirect(self, url):
        """Shortcut all call stack and return response.

        usage: `self.response(url_for(...))`
        """
        raise Redirect(response=redirect(url))


_JSON_HTML = """
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>JSON preview</title>
  <link rel="stylesheet"
    href="{{ url_for('abilian_static', filename="highlightjs/default.min.css") }}" />
</head>
<body>
  <pre>
  <code class="json">
{{ content|escape }}
  </code>
  </pre>

  <script src="{{ url_for('abilian_static',
    filename="highlightjs/highlight.min.js")  }}" ></script>
  <script>hljs.initHighlightingOnLoad();</script>
</body>
</html>
"""


class JSONView(View):
    """Base view for JSON GET.

    Renders as JSON when requested by Ajax, renders as HTML when
    requested from browser.
    """

    def prepare_args(
        self, args: Tuple, kwargs: Dict[Any, Any]
    ) -> Tuple[Tuple, Dict[Any, Any]]:
        kwargs.update({k: v for k, v in request.args.items()})
        return args, kwargs

    def data(self, *args, **kwargs) -> Dict:
        """This method should return data to be serialized using JSON."""
        raise NotImplementedError

    def get(self, *args: Any, **kwargs: Any) -> str:
        data = self.data(*args, **kwargs)
        best_mime = request.accept_mimetypes.best_match(
            ["text/html", "application/json"]
        )
        if best_mime == "application/json":
            return jsonify(data)

        # dev requesting from browser? serve html, let debugtoolbar show up,
        # etc...
        content = json.dumps(data, indent=2)
        return render_template_string(_JSON_HTML, content=content)



from typing import Iterator

from pytest import fixture, mark
from sqlalchemy.orm import Session

from abilian.app import Application
from abilian.core.entities import Entity
from abilian.core.models.subjects import Group, User, create_root_user
from abilian.core.sqlalchemy import SQLAlchemy
from abilian.services.security.models import FolderishModel

from . import (
    READ,
    WRITE,
    Admin,
    Anonymous,
    Authenticated,
    Creator,
    Owner,
    Permission,
    PermissionAssignment,
    Reader,
    Role,
    RoleAssignment,
    SecurityAudit,
    Writer,
    security,
)


def test_singleton() -> None:
    admin = Role("admin")
    other_admin = Role("admin")
    assert admin is other_admin
    assert id(admin) == id(other_admin)


def test_equality() -> None:
    admin = Role("admin")
    assert admin == "admin"


def test_ordering() -> None:
    roles = sorted([Authenticated, Admin, Anonymous])
    assert roles == [Admin, Anonymous, Authenticated]


# TODO: may fail depending on test order, due to global state
@mark.skip
def test_enumerate_assignables(db):
    assert Role.assignable_roles() == [Admin]


@fixture
def session(app: Application, db: SQLAlchemy) -> Iterator[Session]:
    security.start()
    create_root_user()

    yield db.session

    security.stop()
    security.clear()


class DummyModel(Entity):
    pass


def test_anonymous_user(app: Application, session: Session) -> None:
    # anonymous user is not an SQLAlchemy instance and must be handled
    # specifically to avoid tracebacks
    anon = app.login_manager.anonymous_user()
    assert not security.has_role(anon, "reader")
    assert security.has_role(anon, Anonymous)
    assert security.get_roles(anon) == [Anonymous]
    assert not security.has_permission(anon, "read")


def test_has_role_authenticated(app: Application, session: Session) -> None:
    anon = app.login_manager.anonymous_user()
    user = User(email="john@example.com", password="x")
    session.add(user)
    session.flush()
    assert not security.has_role(anon, Authenticated)
    assert security.has_role(user, Authenticated)


def test_root_user(session: Session) -> None:
    # Root user always has any role, any permission.
    root = User.query.get(0)
    assert isinstance(root, User)
    assert security.has_role(root, Anonymous)
    assert security.has_role(root, Admin)
    assert security.has_permission(root, "manage")

    obj = DummyModel()
    session.add(obj)
    session.flush()
    assert security.has_role(root, Admin, obj)
    assert security.has_permission(root, "manage", obj)


def test_grant_basic_roles(session: Session) -> None:
    user = User(email="john@example.com", password="x")
    session.add(user)
    session.flush()

    # everybody always has role Anonymous
    assert security.has_role(user, Anonymous)

    security.grant_role(user, Admin)
    assert security.has_role(user, Admin)
    assert security.get_roles(user) == [Admin]
    assert security.get_roles(user) == ["admin"]
    assert security.get_principals(Admin) == [user]

    # clear roles cache for better coverage: has_permission uses
    # _fill_role_cache_batch(), get_roles uses _fill_role_cache()
    delattr(user, "__roles_cache__")
    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert security.has_permission(user, "manage")

    security.ungrant_role(user, "admin")
    assert not security.has_role(user, "admin")
    assert security.get_roles(user) == []
    assert security.get_principals(Admin) == []

    assert not security.has_permission(user, "read")
    assert not security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")


def test_grant_basic_roles_on_groups(session: Session) -> None:
    user = User(email="john@example.com", password="x")
    group = Group(name="Test Group")
    user.groups.add(group)
    session.add(user)
    session.flush()

    security.grant_role(group, "admin")
    assert security.has_role(group, "admin")
    assert security.get_roles(group) == ["admin"]
    assert security.get_principals(Admin) == [group]

    assert security.has_role(user, Admin)

    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert security.has_permission(user, "manage")

    security.ungrant_role(group, "admin")
    assert not security.has_role(group, "admin")
    assert security.get_roles(group) == []
    assert security.get_principals(Admin) == []

    assert not security.has_role(user, "admin")
    assert not security.has_permission(user, "read")
    assert not security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")


def test_grant_roles_on_objects(session: Session) -> None:
    user = User(email="john@example.com", password="x")
    user2 = User(email="papa@example.com", password="p")
    group = Group(name="Test Group")
    user.groups.add(group)
    obj = DummyModel()
    session.add_all([user, user2, obj])
    session.flush()

    security.grant_role(user, "global_role")
    security.grant_role(user, "reader", obj)
    assert security.has_role(user, "reader", obj)
    assert security.get_roles(user, obj) == ["reader"]
    assert security.get_principals(Reader) == []
    assert security.get_principals(Reader, object=obj) == [user]

    assert security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    # test get_roles "global": object roles should not appear
    assert security.get_roles(user) == ["global_role"]

    # global role is valid on all object
    assert security.has_role(user, "global_role", obj)

    security.ungrant_role(user, "reader", obj)
    assert not security.has_role(user, "reader", obj)
    assert security.get_roles(user, obj) == []
    assert security.has_role(user, "global_role", obj)

    assert not security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    # owner / creator roles
    assert security.get_principals(Owner, object=obj) == []
    assert security.get_principals(Creator, object=obj) == []
    old_owner = obj.owner
    old_creator = obj.creator
    obj.owner = user
    assert security.get_roles(user, obj) == [Owner]
    assert security.get_principals(Owner, object=obj) == [user]
    assert security.get_principals(Creator, object=obj) == []
    # if user2 has Admin role e gets the rights no matter Creator/Ownership
    security.grant_role(user2, Admin)
    assert security.has_role(user2, (Owner, Creator), obj)
    assert security.has_role(user, (Owner, Creator), obj)

    obj.owner = old_owner
    obj.creator = user
    assert security.get_roles(user, obj) == [Creator]
    assert security.get_principals(Owner, object=obj) == []
    assert security.get_principals(Creator, object=obj) == [user]
    obj.creator = old_creator

    # permissions through group membership
    security.grant_role(group, "manager", obj)
    assert security.has_role(group, "manager", obj)
    assert security.get_roles(group, obj) == ["manager"]

    # group membership: user hasn't role set, but has permissions
    assert security.get_roles(user, obj, no_group_roles=True) == []
    assert security.has_permission(user, "read", obj)
    assert security.has_permission(user, "write", obj)
    assert security.has_permission(user, "manage", obj)

    group.members.remove(user)
    session.flush()
    assert not security.has_role(user, "manager", obj)
    assert security.get_roles(user, obj) == []
    assert not security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    security.ungrant_role(group, "manager", obj)
    assert not security.has_role(group, "manager", obj)
    assert security.get_roles(group, obj) == []

    # when called on unmapped instance
    new_obj = DummyModel()
    assert not security.has_permission(user, READ, new_obj)


def test_grant_roles_unique(session: Session) -> None:
    user = User(email="john@example.com", password="x")
    obj = DummyModel()
    session.add_all([user, obj])
    session.flush()

    assert RoleAssignment.query.count() == 0

    security.grant_role(user, "manager", obj)
    session.flush()
    assert RoleAssignment.query.count() == 1

    security.grant_role(user, "manager", obj)
    session.flush()
    assert RoleAssignment.query.count() == 1

    security.grant_role(user, "reader", obj)
    session.flush()
    assert RoleAssignment.query.count() == 2


def test_inherit(session: Session) -> None:
    folder = FolderishModel()
    session.add(folder)
    session.flush()
    assert SecurityAudit.query.count() == 0

    security.set_inherit_security(folder, False)
    session.flush()
    assert not folder.inherit_security
    assert SecurityAudit.query.count() == 1

    security.set_inherit_security(folder, True)
    session.flush()
    assert folder.inherit_security
    assert SecurityAudit.query.count() == 2


def test_add_list_delete_permissions(session: Session) -> None:
    obj = DummyModel()
    assert security.get_permissions_assignments(obj) == {}
    session.add(obj)
    session.flush()

    security.add_permission(READ, Authenticated, obj)
    security.add_permission(READ, Owner, obj)
    security.add_permission(WRITE, Owner, obj)
    assert security.get_permissions_assignments(obj) == {
        READ: {Authenticated, Owner},
        WRITE: {Owner},
    }

    security.delete_permission(READ, Authenticated, obj)
    assert security.get_permissions_assignments(obj) == {READ: {Owner}, WRITE: {Owner}}
    assert security.get_permissions_assignments(obj, READ) == {READ: {Owner}}

    # do it twice: it should not crash
    security.add_permission(READ, Owner, obj)
    security.delete_permission(READ, Authenticated, obj)

    # set/get/delete global permission
    security.add_permission(READ, Writer)
    assert security.get_permissions_assignments() == {READ: {Writer}}


def test_has_permission_on_objects(session: Session) -> None:
    has_permission = security.has_permission
    user = User(email="john@example.com", password="x")
    group = Group(name="Test Group")
    user.groups.add(group)
    obj = DummyModel(creator=user, owner=user)
    session.add_all([user, obj])
    session.flush()

    # global role provides permissions on any object
    security.grant_role(user, Reader)
    assert has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    security.grant_role(user, Writer, obj=obj)
    assert has_permission(user, WRITE, obj=obj)

    # permission assignment
    security.ungrant_role(user, Reader)
    security.ungrant_role(user, Writer, object=obj)
    security.grant_role(user, Authenticated)
    assert not has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    pa = PermissionAssignment(role=Authenticated, permission=READ, object=obj)
    session.add(pa)
    session.flush()
    assert has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    session.delete(pa)
    session.flush()
    assert not has_permission(user, READ, obj=obj)

    # Owner / Creator
    for role in (Owner, Creator):
        pa = PermissionAssignment(role=role, permission=READ, object=obj)
        session.add(pa)
        session.flush()
        assert has_permission(user, READ, obj=obj)

        session.delete(pa)
        session.flush()
        assert not has_permission(user, READ, obj=obj)

    # test when object is *not* in session (newly created objects have id=None
    # for instance)
    obj = DummyModel()
    assert security.has_role(user, Reader, object=obj) is False


def test_has_permission_custom_roles(session: Session) -> None:
    user = User(email="john@example.com", password="x")
    session.add(user)
    session.flush()

    role = Role("custom_role")
    permission = Permission("custom permission")
    assert not security.has_permission(user, permission, roles=role)
    security.grant_role(user, role)
    assert not security.has_permission(user, permission)
    assert security.has_permission(user, permission, roles=role)

    # Permission always granted if Anonymous role
    assert security.has_permission(user, permission, roles=Anonymous)

    # test convert legacy permission & implicit mapping
    security.grant_role(user, "reader")
    assert security.has_permission(user, "read")
    assert not security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")

    security.grant_role(user, "writer")
    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")

    security.grant_role(user, "manager")
    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert security.has_permission(user, "manage")


@mark.skip
def test_query_entity_with_permission(session):
    get_filter = security.query_entity_with_permission
    user = User(email="john@example.com", password="x")
    session.add(user)

    obj_reader = DummyModel(name="reader")
    obj_writer = DummyModel(name="writer")
    obj_none = DummyModel(name="none")
    session.add_all([obj_reader, obj_writer, obj_none])

    assigments = [
        PermissionAssignment(role=Reader, permission=READ, object=obj_reader),
        #
        PermissionAssignment(role=Writer, permission=WRITE, object=obj_writer),
    ]
    session.add_all(assigments)
    session.flush()

    # very unfiltered query returns all objects
    base_query = DummyModel.query
    assert set(base_query.all()) == {obj_reader, obj_writer, obj_none}

    # user has no roles: no objects returned at all
    assert base_query.filter(get_filter(READ, user=user)).all() == []
    assert base_query.filter(get_filter(WRITE, user=user)).all() == []

    # grant object specific roles
    security.grant_role(user, Reader, obj=obj_reader)
    security.grant_role(user, Writer, obj=obj_writer)
    session.flush()
    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # Permission granted to anonymous: objects returned
    pa = PermissionAssignment(role=Anonymous, permission=WRITE, object=obj_reader)
    session.add(pa)

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert set(base_query.filter(get_filter(WRITE, user=user)).all()) == {
        obj_reader,
        obj_writer,
    }
    session.delete(pa)
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # grant global roles
    security.ungrant_role(user, Reader, object=obj_reader)
    security.ungrant_role(user, Writer, object=obj_writer)
    security.grant_role(user, Reader)
    security.grant_role(user, Writer)
    session.flush()

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # admin role has all permissions
    # 1: local role
    security.ungrant_role(user, Reader)
    security.ungrant_role(user, Writer)
    security.grant_role(user, Admin, obj=obj_reader)
    security.grant_role(user, Admin, obj=obj_none)
    session.flush()

    assert set(base_query.filter(get_filter(READ, user=user)).all()) == {
        obj_reader,
        obj_none,
    }
    assert set(base_query.filter(get_filter(WRITE, user=user)).all()) == {
        obj_reader,
        obj_none,
    }

    # 2: global role
    security.ungrant_role(user, Admin, object=obj_reader)
    security.ungrant_role(user, Admin, object=obj_none)
    security.grant_role(user, Admin)
    session.flush()
    assert set(base_query.filter(get_filter(READ, user=user)).all()) == {
        obj_reader,
        obj_writer,
        obj_none,
    }

    # implicit role: Owner, Creator
    security.ungrant_role(user, Admin)
    assert base_query.filter(get_filter(READ, user=user)).all() == []
    assert base_query.filter(get_filter(WRITE, user=user)).all() == []

    obj_reader.creator = user
    obj_writer.owner = user
    assigments = [
        PermissionAssignment(role=Creator, permission=READ, object=obj_reader),
        #
        PermissionAssignment(role=Owner, permission=WRITE, object=obj_writer),
    ]
    session.add_all(assigments)
    session.flush()

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]


def test_add_delete_permissions_expunged_obj(session: Session) -> None:
    # weird case. In CreateObject based views, usually Entity is
    # instanciated and might be added to session if it has a relationship
    # with an existing object.
    # `init_object` must do `session.expunge(obj)`. But entities will
    # have initialized default permissions during `after_attach`.
    #
    # At save time, the object is added again to session. The bug is that
    # without precaution we may create permissions assignment twice, because
    # assignments created in the first place are not yet again in session
    # (new, dirty, deleted) and cannot be found with a filtered query on
    # PermissionAssignment because they have not been flushed yet.
    #
    security.add_permission(READ, Owner, None)
    obj = DummyModel()
    # override default permission at instance level
    obj.__default_permissions__ = frozenset({(READ, frozenset({Owner}))})
    # core.entities._setup_default_permissions creates
    session.add(obj)
    # permissions
    security.add_permission(READ, Owner, obj)
    # no-op obj and its permissions are removed from session
    session.expunge(obj)

    session.add(obj)
    # obj in session again. When
    # _setup_default_permissions is called during
    # `after_flush`, previously created permission are not
    # yet back in session. The cascading rule will add
    # them just after (as of sqlalchemy 0.8, at least)

    # Finally the test! IntegrityError will be raised if we have done
    # something wrong (`Key (permission, role, object_id)=(..., ..., ...)
    # already exists`)
    session.flush()



"""Vocabularies service."""
import typing
from inspect import isclass

import jinja2

from abilian.services.base import Service

from .models import BaseVocabulary

if typing.TYPE_CHECKING:
    from abilian.app import Application


def _vocabularies():
    return {
        cls
        for cls in BaseVocabulary._decl_class_registry.values()
        if isclass(cls) and issubclass(cls, BaseVocabulary)
    }


def _grouped_vocabularies():
    by_group = {}
    for voc in sorted(_vocabularies(), key=lambda v: v.Meta.name):
        by_group.setdefault(voc.Meta.group, []).append(voc)
    return by_group


def get_vocabulary(name: str, group: str = "") -> None:
    name = name.lower()
    vocs = _grouped_vocabularies()
    for voc in vocs.get(group, ()):
        if voc.Meta.name == name:
            return voc

    return None


class VocabularyService(Service):
    name = "vocabularies"

    def init_app(self, app: "Application") -> None:
        super().init_app(app)
        app.register_jinja_loaders(jinja2.PackageLoader(__name__))

    @property
    def vocabularies(self):
        return _vocabularies()

    @property
    def grouped_vocabularies(self):
        return _grouped_vocabularies()

    def get_vocabulary(self, name: str, group: str = "") -> None:
        return get_vocabulary(name, group=group)


vocabularies = VocabularyService()



""""""
from typing import Dict

from flask import current_app
from werkzeug.exceptions import BadRequest

from abilian.core.entities import Entity
from abilian.core.models.tag import Tag
from abilian.i18n import _l
from abilian.web import url_for
from abilian.web.blueprints import Blueprint
from abilian.web.views import (
    BaseObjectView,
    JSONView,
    ObjectCreate,
    ObjectDelete,
    ObjectEdit,
)

from .forms import TagForm

bp = Blueprint("tags", __name__, url_prefix="/tags", template_folder="templates")


class BaseTagView:
    """Mixin for tag views."""

    Model = Tag
    Form = TagForm

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extension = current_app.extensions["tags"]


class TagEdit(BaseTagView, ObjectEdit):
    _message_success = _l("Tag edited")


edit_view = TagEdit.as_view("edit")
bp.route("/manage/<int:object_id>/edit")(edit_view)


class TagCreate(BaseTagView, ObjectCreate):
    _message_success = _l("Tag created")


create_view = TagCreate.as_view("create")
bp.route("/manage/new")(create_view)


class TagDelete(BaseTagView, ObjectDelete):
    _message_success = _l("Tag deleted")


delete_view = TagDelete.as_view("delete")
bp.route("/manage/<int:object_id>/delete")(delete_view)

# Tags on entities
entity_bp = Blueprint("entity_tags", __name__, url_prefix="/tags/entity")


class BaseEntityTagView(BaseTagView):
    def init_object(self, args, kwargs):
        args, kwargs = super().init_object(args, kwargs)
        entity_id = kwargs.pop("entity_id", None)

        if entity_id is not None:
            self.entity = Entity.query.get(entity_id)

        if self.entity is None:
            raise BadRequest("No entity provided")

        return args, kwargs

    def view_url(self):
        return url_for(self.entity)

    def index_url(self):
        return self.view_url()


class EntityTagList(BaseEntityTagView, BaseObjectView, JSONView):
    def get(self, *args, **kwargs):
        return JSONView.get(self, *args, **kwargs)

    def data(self, *args, **kwargs) -> Dict:
        tags = sorted(self.extension.entity_tags(self.entity))
        return {"result": tags}


entity_bp.route("/<int:object_id>/list")(EntityTagList.as_view("list"))


class EntityTagManage(BaseEntityTagView, ObjectEdit):
    methods = ["POST"]

    # operation: add or remove
    mode = None

    def __init__(self, mode, *args, **kwargs):
        super(EntityTagManage, *args, **kwargs)
        assert mode in ("add", "remove")
        self.mode = mode

    def form_valid(self, redirect_to=None):
        ns = self.form.ns.data
        label = self.form.label.data
        op = getattr(self.extension, self.mode)
        op(self.entity, ns=ns, label=label)


entity_bp.route("/<int:object_id>/add")(EntityTagManage.as_view("add", mode="add"))
entity_bp.route("/<int:object_id>/remove")(
    EntityTagManage.as_view("remove", mode="remove")
)


class EntityTagEdit(ObjectEdit):
    Model = Entity

    def init_object(self, args, kwargs):
        args, kwargs = super().init_object(args, kwargs)
        extension = current_app.extensions["tags"]
        self.Form = extension.entity_tags_form(self.obj)
        return args, kwargs

    def view_url(self):
        return url_for(self.obj)

    def index_url(self):
        return self.view_url()


entity_bp.route("/<int:object_id>/edit")(EntityTagEdit.as_view("edit"))



"""Reusable widgets to be included in views.

NOTE: code is currently quite messy. Needs to be refactored.
"""
import base64
import html
import logging
import re
from collections import namedtuple
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Text, Tuple, Union
from urllib import parse

import bleach
import sqlalchemy as sa
import sqlalchemy.orm
import wtforms
from flask import current_app, g, json, render_template, render_template_string
from flask_babel import format_date, format_datetime, format_number, get_locale
from flask_login import current_user
from flask_wtf.file import FileField
from markupsafe import Markup
from sqlalchemy.orm.mapper import Mapper
from wtforms.fields import Field, IntegerField, Label, StringField
from wtforms.form import Form
from wtforms.widgets import HTMLString, Input
from wtforms.widgets import PasswordInput as BasePasswordInput
from wtforms.widgets import Select
from wtforms.widgets import TextArea as BaseTextArea
from wtforms.widgets import html_params
from wtforms_alchemy import ModelFieldList

from abilian.core.entities import Entity
from abilian.core.models.blob import Blob
from abilian.i18n import _, _l
from abilian.services import image
from abilian.services.image import get_format
from abilian.web import url_for
from abilian.web.filters import babel2datepicker, labelize

from .util import babel2datetime

logger = logging.getLogger(__name__)

__all__ = [
    "linkify_url",
    "text2html",
    "Column",
    # Views
    "BaseTableView",
    "MainTableView",
    "RelatedTableView",
    "AjaxMainTableView",
    "SingleView",
    "Panel",
    "Row",
    # Widgets
    "BooleanWidget",
    "Chosen",
    "DateInput",
    "DateTimeInput",
    "DateTimeWidget",
    "DateWidget",
    "DefaultViewWidget",
    "EmailWidget",
    "EntityWidget",
    "FileInput",
    "FloatWidget",
    "ImageInput",
    "ListWidget",
    "ModelListWidget",
    "MoneyWidget",
    "RichTextWidget",
    "Select2",
    "Select2Ajax",
    "TabularFieldListWidget",
    "TagInput",
    "TextArea",
    "URLWidget",
]


def linkify_url(value: str) -> str:
    """Tranform an URL pulled from the database to a safe HTML fragment."""

    value = value.strip()

    rjs = r"[\s]*(&#x.{1,7})?".join(list("javascript:"))
    rvb = r"[\s]*(&#x.{1,7})?".join(list("vbscript:"))
    re_scripts = re.compile(f"({rjs})|({rvb})", re.IGNORECASE)

    value = re_scripts.sub("", value)

    url = value
    if not url.startswith("http://") and not url.startswith("https://"):
        url = f"http://{url}"

    url = parse.urlsplit(url).geturl()
    if '"' in url:
        url = url.split('"')[0]
    if "<" in url:
        url = url.split("<")[0]

    if value.startswith("http://"):
        value = value[len("http://") :]
    elif value.startswith("https://"):
        value = value[len("https://") :]

    if value.count("/") == 1 and value.endswith("/"):
        value = value[0:-1]

    return f'<a href="{url}">{value}</a>&nbsp;<i class="fa fa-external-link"></i>'


def text2html(text: str) -> Union[Markup, str]:
    text = text.strip()
    if re.search("<(p|br)>", text.lower()):
        return text
    if "\n" not in text:
        return text

    lines = text.split("\n")
    lines = [line for line in lines if line]
    paragraphs = [f"<p>{line}</p>" for line in lines]
    return Markup(bleach.clean("\n".join(paragraphs), tags=["p"]))


class Column:
    def __init__(self, **kw):
        for k, w in kw.items():
            setattr(self, k, w)


class View:
    options: Dict[str, Any] = {}

    def get_templates(
        self, specific_template_name: str, default_template: str
    ) -> List[str]:
        specific_template = self.options.get(specific_template_name)
        if specific_template:
            return [specific_template, default_template]
        else:
            return [default_template]


# TODO: rewrite
class BaseTableView(View):
    show_controls = False
    show_search = None
    paginate = False

    columns: List[Dict[str, Any]]

    def __init__(
        self, columns: List[Dict[str, Any]], options: Optional[Any] = None
    ) -> None:
        if self.show_search is None:
            self.show_search = self.show_controls

        self.columns = []
        self.init_columns(columns)
        self.name = f"{self.__class__.__name__.lower()}-{next(g.id_generator):d}"
        if options is not None:
            self.options = options
            self.show_controls = self.options.get("show_controls", self.show_controls)
            self.show_search = self.options.get("show_search", self.show_controls)
            self.paginate = self.options.get("paginate", self.paginate)

    def init_columns(self, columns: Sequence[Union[str, Dict[str, Any]]]) -> None:
        # TODO
        default_width = f"{(0.99 / len(columns) * 100):2.0f}%"
        for col in columns:
            if isinstance(col, str):
                column = {"name": col, "width": default_width}
            else:
                column = col
            assert isinstance(column, dict)
            column.setdefault("width", default_width)
            column.setdefault("sorting", ("asc", "desc"))
            if "label" not in column:
                column["label"] = labelize(column["name"])
            self.columns.append(column)

    def render(self, entities: List[Entity], **kwargs: Any) -> Markup:
        aoColumns = []
        aaSorting = []
        offset = 0
        if self.show_controls:
            aoColumns.append({"asSorting": []})
            offset = 1

        for idx, c in enumerate(self.columns, offset):
            aoColumns.append({"asSorting": c["sorting"], "sWidth": str(c["width"])})
            aaSorting.append([idx, c["sorting"][0]])

        datatable_options = {
            "aaSorting": aaSorting,
            "aoColumns": aoColumns,
            "bFilter": self.show_search,
            "oLanguage": {
                "sSearch": self.options.get("search_label", _("Filter records:"))
            },
            "bStateSave": False,
            "bPaginate": self.paginate,
            "sPaginationType": "bootstrap",
            "bLengthChange": False,
            "iDisplayLength": self.options.get("paginate_length", 50),
        }
        js = render_template_string(
            """
            require(
                ['jquery', 'jquery.dataTables'],
                function($) {
                    $('#{{ table_id  }}').dataTable({{ options|tojson|safe }});
                });
            """,
            table_id=self.name,
            options=datatable_options,
        )

        table = []
        for entity in entities:
            table.append(self.render_line(entity))

        templates = self.get_templates("template", "widgets/render_table.html")
        return Markup(
            render_template(templates, table=table, js=Markup(js), view=self, **kwargs)
        )

    def render_line(self, entity: Entity) -> List[Union[Markup, str]]:
        line = []
        make_link_on = self.options.get("make_link_on")

        for col in self.columns:
            if isinstance(col, str):
                column_name = col
                build_url = url_for
            else:
                column_name = col["name"]
                build_url = col.get("url", url_for)

            value = entity
            attr = column_name.rsplit(".", 1)
            if len(attr) > 1:
                try:
                    value = getattr(value, attr[0], "")
                    value = value.display_value(attr[1])
                except AttributeError:
                    value = ""
            else:
                value = value.display_value(attr[0])

            format = col.get("format")
            if format:
                cell = format(value)
            elif column_name in (make_link_on, "name") or col.get("linkable"):
                url = build_url(entity)
                text = html.escape(str(value))
                cell = Markup(f'<a href="{url}">{text}</a>')
            elif isinstance(value, Entity):
                url = build_url(entity)
                text = html.escape(value.name)
                cell = Markup(f'<a href="{url}">{text}</a>')
            elif isinstance(value, str) and (
                value.startswith("http://") or value.startswith("www.")
            ):
                cell = Markup(linkify_url(value))
            elif value in (True, False):
                cell = "\u2713" if value else ""  # Unicode "Check mark"
            elif isinstance(value, list):
                cell = "; ".join(value)
            else:
                if not isinstance(value, (Markup,) + (str,)):
                    if value is None:
                        value = ""
                    else:
                        value = str(value)
                cell = value

            line.append(cell)
        return line


class MainTableView(BaseTableView):
    """Table view for main objects list."""

    show_controls = True
    paginate = True


class RelatedTableView(BaseTableView):
    """Table view for related objects list."""

    show_controls = False
    paginate = False


class AjaxMainTableView(View):
    """Variant of the MainTableView that gets content from AJAX requests.

    TODO: refactor all of this (currently code is copy/pasted!).
    """

    show_controls = False
    paginate = True
    columns: List[Dict[str, Any]]
    options: Dict[str, Any]

    def __init__(
        self,
        columns: List[Dict[str, Any]],
        ajax_source: str,
        search_criterions: Tuple = (),
        name: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.columns = []
        self.init_columns(columns)
        self.ajax_source = ajax_source
        self.search_criterions = search_criterions
        self.name = name if name is not None else id(self)
        self.save_state = name is not None
        self.options = options or {}

    def init_columns(self, columns: Sequence[Union[str, Dict[str, Any]]]) -> None:
        # TODO: compute the correct width for each column.
        default_width = 0.99 / len(columns)
        for col in columns:
            if isinstance(col, str):
                column = {"name": col, "width": default_width}
            else:
                column = col
            assert isinstance(column, dict)
            if "label" not in column:
                column["label"] = labelize(column["name"])

            column.setdefault("sorting", ["asc", "desc"])

            if not column["sorting"]:
                column.setdefault("sortable", False)
            else:
                column.setdefault("sortable", True)

            self.columns.append(column)

    def render(self) -> Markup:
        aoColumns = [{"asSorting": []}] if self.show_controls else []
        aoColumns += [
            {"asSorting": col["sorting"], "bSortable": col["sortable"]}
            for col in self.columns
        ]
        datatable_options = {
            "sDom": "fFriltip",
            "aoColumns": aoColumns,
            "bFilter": True,
            "oLanguage": {
                "sSearch": self.options.get("search_label", _("Filter records:")),
                "oPaginate": {"sPrevious": _("Previous"), "sNext": _("Next")},
                "sLengthMenu": _("Entries per page: _MENU_"),
                "sInfo": _("Showing _START_ to _END_ of _TOTAL_ entries"),
                "sInfoEmpty": _("Showing _START_ to _END_ of _TOTAL_ entries"),
                "sInfoFiltered": _("(filtered from _MAX_ total entries)"),
                "sAddAdvancedFilter": _("Add a filter"),
                "sZeroRecords": _("No matching records found"),
                "sEmptyTable": _("No matching records found"),
            },
            "bPaginate": self.paginate,
            "sPaginationType": "bootstrap",
            "bLengthChange": True,
            "iDisplayLength": 25,
            "bStateSave": self.save_state,
            "bProcessing": True,
            "bServerSide": True,
            "sAjaxSource": self.ajax_source,
        }
        if self.options.get("aaSorting", None):
            datatable_options["aaSorting"] = self.options.get("aaSorting")

        advanced_search_filters = []
        for criterion in self.search_criterions:
            if not criterion.has_form_filter:
                continue
            d = {
                "name": criterion.name,
                "label": str(criterion.label),
                "type": criterion.form_filter_type,
                "args": criterion.form_filter_args,
                "unset": criterion.form_unset_value,
            }
            if criterion.has_form_default_value:
                d["defaultValue"] = criterion.form_default_value

            advanced_search_filters.append(d)

        if advanced_search_filters:
            datatable_options["aoAdvancedSearchFilters"] = advanced_search_filters

        return Markup(
            render_template(
                "widgets/render_ajax_table.html",
                datatable_options=datatable_options,
                view=self,
            )
        )

    def render_line(self, entity):
        line = []
        for col in self.columns:
            if isinstance(col, str):
                column_name = col
            else:
                column_name = col["name"]

            value = entity.display_value(column_name)
            has_custom_display = False

            # Manual massage.
            # 'display_fun' gets value *and* entity: useful to perform
            # specific markup based on other entity values.
            # 'display_fmt' is for simple value formatting (format_date from
            # babel for example)
            if value is None:
                value = ""
            elif "display_fun" in col:
                has_custom_display = True
                value = col["display_fun"](entity, value)
            elif "display_fmt" in col:
                value = col["display_fmt"](value)

            if has_custom_display:
                cell = value
            elif column_name == "name":
                cell = Markup(f'<a href="{url_for(entity)}">{html.escape(value)}</a>')
            elif isinstance(value, Entity):
                cell = Markup(
                    f'<a href="{url_for(value)}">{html.escape(value.name)}</a>'
                )
            elif isinstance(value, str) and (
                value.startswith("http://") or value.startswith("www.")
            ):
                cell = Markup(linkify_url(value))
            elif col.get("linkable"):
                cell = Markup(
                    f'<a href="{url_for(entity)}">{html.escape(str(value))}</a>'
                )
            else:
                cell = str(value)

            line.append(cell)
        return line


#
# Single object view
#
class SingleView(View):
    """View on a single object."""

    def __init__(self, form: Form, *panels: "Panel", **options: Any) -> None:
        self.form = form
        self.panels = panels
        self.options = options

    def render(self, item: Entity, related_views: Tuple[()] = ()) -> Markup:
        mapper = sa.orm.class_mapper(item.__class__)
        panels = []
        _to_skip = (None, False, 0, 0.0, "", "-")

        for panel in self.panels:
            data = {}
            field_name_iter = (fn for row in panel.rows for fn in row)

            for name in field_name_iter:
                field = self.form._fields[name]
                if field.is_hidden:
                    continue

                value = field.data
                if not isinstance(field, FileField) and not field.flags.render_empty:
                    if value in _to_skip:
                        continue

                value = Markup(field.render_view(entity=item))
                if value == "":
                    # related models may have [] as value, but we don't discard this type
                    # of value in order to let widget a chance to render something useful
                    # like an 'add model' button.
                    #
                    # if it renders an empty string, there's really no point in rendering
                    # a line for this empty field
                    continue

                label = self.label_for(field, mapper, name)
                data[name] = (label, value)

            if data:
                panels.append((panel, data))

        templates = self.get_templates("view_template", "widgets/render_single.html")
        ctx = {
            "view": self,
            "related_views": related_views,
            "entity": item,
            "panels": panels,
            "form": self.form,
        }
        return Markup(render_template(templates, **ctx))

    def render_form(self, for_new=False, has_save_and_add_new=False):
        # Client-side rules for jQuery.validate
        # See: http://docs.jquery.com/Plugins/Validation/Methods#Validator
        rules = {}
        for field in self.form:
            rules_for_field = {}
            for validator in field.validators:
                rule_for_validator = getattr(validator, "rule", None)
                if rule_for_validator:
                    rules_for_field.update(rule_for_validator)
            if rules_for_field:
                rules[field.name] = rules_for_field
        if rules:
            rules = Markup(json.dumps(rules))
        else:
            rules = None

        templates = self.get_templates("edit_template", "widgets/render_for_edit.html")
        ctx = {
            "view": self,
            "form": self.form,
            "for_new": for_new,
            "has_save_and_add_new": has_save_and_add_new,
            "rules": rules,
        }
        return Markup(render_template(templates, **ctx))

    def label_for(self, field: Field, mapper: Mapper, name: str) -> Label:
        label = field.label
        if label is None:
            try:
                info = mapper.c[name].info
                label = info["label"]
            except (AttributeError, KeyError):
                pass

        if label is None:
            try:
                label = _(name)
            except KeyError:
                # i18n may be not initialized (in some unit tests for example)
                label = name

        return label


#
# Used to describe single entity views.
#
class Panel:
    """`Panel` and `Row` classes help implement a trivial internal DSL for
    specifying multi-column layouts in forms or object views.

    They are currently not really used, since we went with 1-column
    designs eventually.
    """

    def __init__(self, label: str, *rows: "Row") -> None:
        self.label = label
        self.rows = rows

    def __iter__(self):
        return iter(self.rows)

    def __getitem__(self, item):
        return self.rows[item]

    def __len__(self):
        return len(self.rows)


class Row:
    """`Panel` and `Row` classes help implement a trivial internal DSL for
    specifying multi-column layouts in forms or object views.

    They are currently not really used, since we went with 1-column
    designs eventually.
    """

    def __init__(self, *cols: str) -> None:
        self.cols = cols

    def __iter__(self):
        return iter(self.cols)

    def __getitem__(self, item):
        return self.cols[item]

    def __len__(self):
        return len(self.cols)


class ModelWidget:
    edit_template = "widgets/model_widget_edit.html"
    view_template = "widgets/model_widget_view.html"

    def __init__(self, edit_template=None, view_template=None):
        if edit_template is not None:
            self.edit_template = edit_template
        if view_template is not None:
            self.view_template = view_template

    def __call__(self, field, *args, **kwargs):
        return render_template(self.edit_template, form=field)

    def render_view(self, field, *args, **kwargs):
        _to_skip = (None, False, 0, 0.0, "", "-")
        rows = []
        for f in field.form:
            if f.is_hidden:
                continue

            value = f.data
            if value in _to_skip and not f.flags.render_empty:
                continue

            value = Markup(f.render_view())
            if value == "":
                # related models may have [] as value, but we don't discard this type
                # of value in order to let widget a chance to render something useful
                # like an 'add model' button.
                #
                # if it renders an empty string, there's really no point in rendering
                # a line for this empty field
                continue

            label = f.label
            rows.append((label, value))

        return render_template(self.view_template, field=field, rows=rows)


# Form field widgets #####################################################
class TextInput(wtforms.widgets.TextInput):
    """Support pre and post icons.

    An Icon can be a plain string, or an instance of
    :class:`abilian.web.action.Icon`.
    """

    pre_icon: Optional[Text] = None
    post_icon: Optional[Text] = None

    def __init__(
        self,
        input_type: Optional[Any] = None,
        pre_icon: Optional[Any] = None,
        post_icon: Optional[Any] = None,
    ) -> None:
        super().__init__(input_type)

        if pre_icon is not None:
            self.pre_icon = pre_icon
        if post_icon is not None:
            self.post_icon = post_icon

    def __call__(self, field, **kwargs):
        if not any((self.pre_icon, self.post_icon)):
            return super().__call__(field, **kwargs)

        kwargs.setdefault("type", self.input_type)
        if "value" not in kwargs:
            kwargs["value"] = field._value()

        params = self.html_params(name=field.name, **kwargs)
        return Markup(
            render_template_string(
                """
                <div class="input-group input-group-type-{{ widget.typename }}">
                {%- if widget.pre_icon %}
                  <div class="input-group-addon">{{ widget.pre_icon }}</div>
                {%- endif %}
                  <input {{ params | safe}}>
                {%- if widget.post_icon %}
                  <div class="input-group-addon">{{ widget.post_icon }}</div>
                {%- endif %}
                </div>
                """,
                widget=self,
                params=params,
            )
        )

    @property
    def typename(self):
        return self.__class__.__name__.lower()


class TextArea(BaseTextArea):
    """Accepts "resizeable" parameter: "vertical", "horizontal", "both",
    None."""

    _resizeable_valid = ("vertical", "horizontal", "both", None)
    resizeable = None
    rows = None

    def __init__(
        self,
        resizeable: Optional[str] = None,
        rows: Optional[int] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        if resizeable not in self._resizeable_valid:
            raise ValueError(
                "Invalid value for resizeable: {}, valid values are: {!r}"
                "".format(resizeable, self._resizeable_valid)
            )
        if resizeable:
            self.resizeable = f"resizeable-{resizeable}"
        else:
            self.resizeable = "not-resizeable"

        if rows:
            self.rows = int(rows)

    def __call__(self, *args, **kwargs):
        if self.resizeable:
            css = kwargs.get("class_", "")
            kwargs["class_"] = f"{css} {self.resizeable}"

        if self.rows and "rows" not in kwargs:
            kwargs["rows"] = self.rows

        return super().__call__(*args, **kwargs)


class FileInput:
    """Renders a file input.

    Inspired from http://www.abeautifulsite.net/blog/2013/08/whipping-
    file-inputs-into-shape-with-bootstrap-3/
    """

    def __init__(self, template: str = "widgets/file_input.html") -> None:
        self.template = template

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        kwargs["name"] = field.name
        kwargs["type"] = "file"
        kwargs["disabled"] = "disabled"  # JS widget will activate it
        input_elem = f"<input {html_params(**kwargs)}>"
        button_label = _("Add file") if "multiple" in kwargs else _("Select file")

        existing = self.build_exisiting_files_list(field)
        uploads = self.build_uploads_list(field)

        if not field.multiple and uploads:
            # single file field: exising file replaced by new upload, don't show
            # existing
            existing = []
        else:
            for data in existing:
                # this might be a very big bytes object: in any case it will not be used
                # in widget's template, and during development it makes very large pages
                # due to debugtoolbar capturing template parameters
                del data["file"]

        ctx = {
            "id": field.id,
            "field": field,
            "widget": self,
            "input": input_elem,
            "button_label": button_label,
            "existing": existing,
            "uploaded": uploads,
        }
        return Markup(render_template(self.template, **ctx))

    def build_exisiting_files_list(self, field):
        existing = []
        blob = None
        object_data = field.object_data

        if not object_data:
            return existing

        if not field.multiple:
            object_data = [object_data]
            blob = field.blob

        for idx, data in enumerate(object_data):
            if data is not None:
                if hasattr(data, "content_length"):
                    size = data.content_length
                else:
                    try:
                        size = len(bytes(data))
                    except Exception:
                        # FIXME: when does it happen ?
                        size = 0
                existing.append(
                    {
                        "file": data,
                        "size": size,
                        "delete": idx in field.delete_files_index,
                    }
                )
                if hasattr(data, "filename"):
                    existing[-1]["filename"] = data.filename
                elif blob and isinstance(blob, Blob):
                    existing[-1]["filename"] = blob.meta.get("filename", "")

        return existing

    def build_uploads_list(self, field):
        uploads = current_app.extensions["uploads"]
        uploaded = []

        for handle in field.upload_handles:
            file_ = uploads.get_file(current_user, handle)
            if file_ is None:
                continue
            meta = uploads.get_metadata(current_user, handle)
            uploaded.append(
                {
                    "file": file_,
                    "handle": handle,
                    "filename": meta.get("filename", handle),
                    "size": file_.stat().st_size,
                }
            )

        return uploaded


class ImageInput(FileInput):
    """An image widget with client-side preview.

    To show current image field data has to provide an attribute named
    `url`.
    """

    def __init__(
        self,
        template: str = "widgets/image_input.html",
        width: int = 120,
        height: int = 120,
        resize_mode: str = image.CROP,
        valid_extensions: Tuple[str, str, str] = ("jpg", "jpeg", "png"),
    ) -> None:
        super().__init__(template=template)
        self.resize_mode = resize_mode
        self.valid_extensions = valid_extensions
        self.width, self.height = width, height

    def build_exisiting_files_list(self, field):
        existing = super().build_exisiting_files_list(field)
        for data in existing:
            value = data["file"]
            if value:
                if hasattr(value, "url"):
                    image_url = value.url
                else:
                    image_url = self.get_b64_thumb_url(
                        self.get_thumb(value, self.width, self.height)
                    )

                data["image_url"] = image_url

        return existing

    def build_uploads_list(self, field):
        uploaded = super().build_uploads_list(field)
        for data in uploaded:
            value = data["file"]
            if value:
                if hasattr(value, "url"):
                    image_url = value.url
                else:
                    with value.open("rb") as in_:
                        image_url = self.get_b64_thumb_url(
                            self.get_thumb(in_, self.width, self.height)
                        )

                data["image_url"] = image_url

        return uploaded

    def get_thumb(self, data, width, height):
        try:
            get_format(data)
        except OSError:
            return ""
        return image.resize(data, width, height, mode=self.resize_mode)

    def get_b64_thumb_url(self, img):
        if not img:
            return ""
        try:
            fmt = image.get_format(img).lower()
        except OSError:
            return ""
        thumb = base64.b64encode(img).decode("ascii")
        return f"data:image/{fmt};base64,{thumb}"

    def render_view(self, field, **kwargs):
        data = field.data
        if not data:
            return ""
        try:
            get_format(data)
        except OSError:
            return ""

        width = kwargs.get("width", self.width)
        height = kwargs.get("heigth", self.height)

        thumb = self.get_thumb(data, width, height)
        width, height = image.get_size(thumb)

        tmpl = '<img src="{{ url }}" width="{{ width }}" height="{{ height }}" />'
        ctx = {"url": self.get_b64_thumb_url(thumb), "width": width, "height": height}
        return render_template_string(tmpl, **ctx)


class Chosen(Select):
    """Extends the Select widget using the Chosen jQuery plugin."""

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        params = html_params(name=field.name, **kwargs)
        html = [f'<select {params} class="chzn-select">']
        for val, label, selected in field.iter_choices():
            html.append(self.render_option(val, label, selected))
        html.append("</select>")
        return Markup("".join(html))

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        options = dict(kwargs, value=value)
        if selected:
            options["selected"] = True
        params = html_params(**options)
        return Markup(f"<option {params}>{html.escape(str(label))}</option>")


class TagInput(Input):
    """Extends the Select widget using the Chosen jQuery plugin."""

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        kwargs["class_"] = "tagbox"
        if "value" not in kwargs:
            kwargs["value"] = field._value()

        params = self.html_params(name=field.name, **kwargs)
        return Markup(f"<input {params}>")


class DateInput(Input):
    """Renders date inputs using the fancy Bootstrap Datepicker:

    https://github.com/eternicode/bootstrap-datepicker
    """

    input_type = "date"

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        kwargs.setdefault("name", field.name)
        field_name = kwargs.pop("name")
        kwargs.pop("type", None)
        value = kwargs.pop("value", None)
        if value is None:
            value = field._value()
        if not value:
            value = ""

        date_fmt = kwargs.pop("format", None)
        if date_fmt is not None:
            date_fmt = (
                date_fmt.replace("%", "")
                .replace("d", "dd")
                .replace("m", "mm")
                .replace("Y", "yyyy")
            )
        else:
            date_fmt = get_locale().date_formats["short"].pattern
            date_fmt = babel2datepicker(date_fmt)
            date_fmt = date_fmt.replace("M", "m")  # force numerical months

        attributes = {
            "class": "input-group date",
            "data-provide": "datepicker",
            "data-date": value,
            "data-date-format": date_fmt,
            "data-date-autoclose": "true",
        }

        s = f"<div {html_params(**attributes)}>\n"

        s += '  <input size="13" type="text" {} />\n'.format(
            html_params(name=field_name, id=field_id, value=value, **kwargs)
        )
        s += '  <span class="input-group-addon"><i class="fa fa-calendar"></i></span>\n'
        s += "</div>\n"
        return Markup(s)

    def render_view(self, field, **kwargs):
        return format_date(field.data) if field.data else ""


class TimeInput(Input):
    """Renders time inputs using boostrap timepicker:

    https://github.com/jdewit/bootstrap-timepicker
    """

    template = "widgets/timepicker.html"

    def __init__(
        self,
        template: None = None,
        widget_mode: str = "dropdown",
        h24_mode: bool = True,
        minuteStep: int = 1,
        showSeconds: bool = False,
        secondStep: int = 1,
        showInputs: bool = False,
        disableFocus: bool = False,
        modalBackdrop: bool = False,
    ) -> None:
        super().__init__()

        if template is not None:
            self.template = template

        self.widget_mode = widget_mode
        self.h24_mode = h24_mode
        self.minuteStep = minuteStep
        self.showSeconds = showSeconds
        self.secondStep = secondStep
        self.showInputs = showInputs
        self.disableFocus = disableFocus
        self.modalBackdrop = modalBackdrop
        self.strptime = "%H:%M" if self.h24_mode else "%I:%M %p"

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        value = kwargs.pop("value", None)
        if value is None:
            value = field._value()
        if not value:
            value = ""

        time_fmt = get_locale().time_formats["short"].format
        is_h12 = "%(h)s" in time_fmt or "%(K)s" in time_fmt

        input_params = {
            "data-template": self.widget_mode,
            "data-show-meridian": is_h12,
            "data-minute-step": self.minuteStep,
            "data-show-seconds": self.showSeconds,
            "data-second-step": self.secondStep,
            "data-show-inputs": self.showInputs,
            "data-disable-focus": self.disableFocus,
            "data-modal-backdrop": self.modalBackdrop,
        }

        input_params = {k: Markup(json.dumps(v)) for k, v in input_params.items()}

        ctx = {
            "id": field_id,
            "value": value,
            "field": field,
            "required": False,
            "timepicker_attributes": input_params,
        }
        return Markup(render_template(self.template, **ctx))


class DateTimeInput:
    """if corresponding field.raw_data exist it is used to initialize default
    date & time (raw_data example: ["10/10/16 | 09:00"])"""

    def __init__(self) -> None:
        self.date = DateInput()
        self.time = TimeInput()

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        kwargs.setdefault("name", field.name)
        field_name = kwargs.pop("name")

        locale = get_locale()
        date_fmt = locale.date_formats["short"].pattern
        date_fmt = babel2datetime(date_fmt)
        # force numerical months
        date_fmt = date_fmt.replace("%B", "%m").replace("%b", "%m")
        time_fmt = "%H:%M"

        value = kwargs.pop("value", None)
        if value is None:
            if field.data:
                value = field.data
                date_value = value.strftime(date_fmt) if value else ""
                time_value = value.strftime(time_fmt) if value else ""
            elif field.raw_data:
                value = field.raw_data  # example "10/10/16 | 09:00"
                value = "".join(value)
                value = value.split("|")

                try:
                    date_value = datetime.strptime(value[0].strip(), date_fmt)
                except (ValueError, IndexError):
                    date_value = ""

                try:
                    time_value = value[1].strip()  # example  "09:00"
                except IndexError:
                    time_value = ""

            else:
                date_value = ""
                time_value = ""

        return (
            Markup(
                '<div class="form-inline">\n'
                '<input class="datetimepicker" type="hidden" id="{id}" '
                'name="{name}" value="{date} | {time}" />\n'
                "".format(
                    id=field_id, name=field_name, date=date_value, time=time_value
                )
            )
            + self.date(
                field,
                id=f"{field_id}-date",
                name=f"{field_name}-date",
                value=date_value,
            )
            + self.time(
                field,
                id=f"{field_id}-time",
                name=f"{field_name}-time",
                value=time_value,
            )
            + Markup("</div>")
        )


class DefaultViewWidget:
    def render_view(
        self, field: Union[IntegerField, StringField], **kwargs: Any
    ) -> str:
        value = field.object_data
        if isinstance(value, str):
            return text2html(value)
        else:
            # [], None and other must be rendered using empty string
            return str(value or "")


class BooleanWidget(wtforms.widgets.CheckboxInput):
    # valid data-* options when using boostrap-switch
    _ON_OFF_VALID_OPTIONS = frozenset(
        (
            "animate",
            "indeterminate",
            "inverse",
            "radio-all-off",
            "on-color",
            "off-color",
            "on-text",
            "off-text",
            "label-text",
            "handle-width",
            "label-width",
            "base-class",
            "wrapper-class",
        )
    )
    on_off_options: Dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.on_off_mode = kwargs.pop("on_off_mode", False)
        self.on_off_options = {}
        on_off_options = kwargs.pop("on_off_options", {})
        for k, v in on_off_options.items():
            if k not in self._ON_OFF_VALID_OPTIONS:
                continue
            self.on_off_options[f"data-{k}"] = v

        if self.on_off_mode:
            self.on_off_options["data-toggle"] = "on-off"

        super().__init__(*args, **kwargs)

    def __call__(self, field, **kwargs):
        if self.on_off_mode:
            kwargs.update(self.on_off_options)

        return super().__call__(field, **kwargs)

    def render_view(self, field, **kwargs):
        return "\u2713" if field.object_data else ""  # Text_type "Check mark"


class PasswordInput(BasePasswordInput):
    """Supports setting 'autocomplete' at instanciation time."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.autocomplete = kwargs.pop("autocomplete", None)
        super().__init__(*args, **kwargs)

    def __call__(self, field, **kwargs):
        kwargs.setdefault("autocomplete", self.autocomplete)
        return super().__call__(field, **kwargs)

    def render_view(self, field, **kwargs):
        return "*****"


class FloatWidget(wtforms.widgets.TextInput):
    """In view mode, format float number to 'precision' decimal."""

    def __init__(self, precision=None):
        self.precision = precision
        if precision is not None:
            self._fmt = f".{precision:d}f"

    def render_view(self, field, **kwargs):
        data = field.object_data
        if data is None:
            return ""

        return format(data, self._fmt)


class DateWidget(wtforms.widgets.TextInput):
    def render_view(self, field, **kwargs):
        return format_date(field.object_data) if field.object_data else ""


class DateTimeWidget(DateWidget):
    def render_view(self, field, **kwargs):
        return format_datetime(field.object_data) if field.object_data else ""


class EntityWidget:
    def render_view(self, field, **kwargs):
        objs = field.object_data
        if not field.multiple:
            objs = [objs]
        return ", ".join(
            f'<a href="{url_for(o)}">{html.escape(o.name)}</a>' for o in objs if o
        )


class HoursWidget(TextInput):
    """Widget used to show / enter hours.

    Currently hardcoded to heure(s)
    """

    post_icon = _l("hour(s)")
    input_type = "number"

    def render_view(self, field, **kwargs):
        val = field.object_data
        unit = "h"

        if val is None:
            return ""

        # \u00A0: non-breakable whitespace
        return f"{val}\u00A0{unit}"


class MoneyWidget(TextInput):
    """Widget used to show / enter money amount.

    Currently hardcoded to € / k€.
    """

    post_icon = "€"
    input_type = "number"

    def render_view(self, field, **kwargs):
        val = field.object_data
        unit = "€"

        if val is None:
            return ""

        if val > 1000:
            unit = "k€"
            # pylint: disable=W1633
            val = int(round(val / 1000.0))

        # `format_currency()` is not used since it display numbers with cents
        # units, which we don't want
        #
        # \u00A0: non-breakable whitespace
        return f"{format_number(val)} {unit}"


class EmailWidget(TextInput):
    pre_icon = "@"

    def render_view(self, field: StringField, **kwargs: Any) -> str:
        links = ""
        if isinstance(field, wtforms.fields.FieldList):
            for entry in field.entries:
                link = bleach.linkify(entry.data, parse_email=True)
                if link:
                    links += f' {link}&nbsp;<i class="fa fa-envelope"></i><br>'
        else:
            link = bleach.linkify(field.object_data, parse_email=True)
            if link:
                links = f'{link}&nbsp;<i class="fa fa-envelope"></i>'
        return links


class URLWidget:
    def render_view(self, field, **kwargs):
        return linkify_url(field.object_data) if field.object_data else ""


class RichTextWidget:
    template = "widgets/richtext.html"
    allowed_tags = {
        "a": {"href": True, "title": True},
        "abbr": {"title": True},
        "acronym": {"title": True},
        "b": True,
        "blockquote": True,
        "br": True,
        "code": True,
        "em": True,
        "h1": True,
        "h2": True,
        "h3": True,
        "h4": True,
        "h5": True,
        "h6": True,
        "i": True,
        "img": {"src": True},
        "li": True,
        "ol": True,
        "strong": True,
        "ul": True,
        "p": True,
        "u": True,
    }
    profile = "standard"

    def __init__(self, allowed_tags=None, template=None, profile=None):
        if allowed_tags is not None:
            self.allowed_tags = allowed_tags
        if template is not None:
            self.template = template
        if profile is not None:
            self.profile = profile

    def __call__(self, field, **kwargs):
        value = kwargs.pop("value") if "value" in kwargs else field._value()
        kwargs.setdefault("allowed_tags", self.allowed_tags)
        kwargs.setdefault("profile", self.profile)
        return render_template(self.template, field=field, value=value, kw=kwargs)


class ListWidget(wtforms.widgets.ListWidget):
    """display field label is optionnal."""

    def __init__(
        self, html_tag: str = "ul", prefix_label: bool = True, show_label: bool = True
    ) -> None:
        super().__init__(html_tag, prefix_label)
        self.show_label = show_label

    def __call__(self, field, **kwargs):
        if self.show_label:
            return super().__call__(field, **kwargs)

        kwargs.setdefault("id", field.id)
        html = [f"<{self.html_tag} {wtforms.widgets.html_params(**kwargs)}>"]
        for subfield in field:
            html.append(f"<li>{subfield()}</li>")

        html.append(f"</{self.html_tag}>")
        return Markup("".join(html))

    def render_view(self, field, **kwargs):
        data = field.data
        is_empty = data == [] if field.multiple else data is None

        if not is_empty:
            data = (
                [label for v, label, checked in field.iter_choices() if checked]
                if (hasattr(field, "iter_choices") and callable(field.iter_choices))
                else field.object_data
            )
        else:
            data = []

        tpl = """{%- for obj in data %}
              <span class="badge">{{ obj }}</span>
              {%- endfor %}"""
        return Markup(render_template_string(tpl, data=data))


class FieldListWidget:
    """For list of Fields (using <tr><td>)"""

    view_template = "widgets/fieldlist_view.html"
    template = "widgets/fieldlist.html"

    def __init__(self, template=None, view_template=None):
        if template is not None:
            self.template = template
        if view_template is not None:
            self.view_template = view_template

    def __call__(self, field, **kwargs):
        assert isinstance(field, wtforms.fields.FieldList)
        return Markup(render_template(self.template, field=field))

    def render_view(self, field, **kwargs):
        assert isinstance(field, wtforms.fields.FieldList)
        value = field.object_data
        if not value:
            return ""
        return Markup(render_template(self.view_template, field=field))


class TabularFieldListWidget:
    """For list of formfields.

    2 templates are available:

    * widgets/tabular_fieldlist_widget.html (default):
      Show sub-forms as a table, one row of inputs per model
    * widgets/model_fieldlist.html:
      Show sub-forms as a list of forms
    """

    def __init__(self, template="widgets/tabular_fieldlist_widget.html"):
        self.template = template

    def __call__(self, field, **kwargs):
        assert isinstance(field, wtforms.fields.FieldList)
        labels = None

        if len(field):
            assert isinstance(field[0], wtforms.fields.FormField)
            field_names = [f.short_name for f in field[0] if not f.is_hidden]
            data_type = f"{field.entries[0].__class__.__name__}Data"
            Data = namedtuple(data_type, field_names)
            labels = Data(*[f.label for f in field[0] if not f.is_hidden])

        return Markup(render_template(self.template, labels=labels, field=field))


class ModelListWidget:
    def __init__(self, template="widgets/horizontal_table.html"):
        self.template = template

    def render_view(self, field, **kwargs):
        assert isinstance(field, ModelFieldList)
        value = field.object_data
        if not value:
            return render_template(
                self.template, field=field, labels=(), rows=(), **kwargs
            )

        field_names = field._field_names
        labels = field._field_labels
        data_type = f"{field.entries[0].object_data.__class__.__name__}Data"
        Data = namedtuple(data_type, field_names)
        labels = Data(*labels)

        rows = []
        for entry in field.entries:
            row = []
            for f in filter(lambda f: not f.is_hidden, entry.form):
                row.append(Markup(f.render_view()))

            rows.append(Data(*row))

        rendered = render_template(
            self.template, field=field, labels=labels, rows=rows, **kwargs
        )
        return rendered


#
# Selection widget.
#
class Select2(Select):
    """Transforms a Select widget into a Select2 widget.

    Depends on global JS code.
    """

    unescape_html = False

    def __init__(
        self,
        unescape_html: bool = False,
        js_init: str = "select2",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.js_init = js_init

        if unescape_html:
            self.unescape_html = True

    def __call__(self, field, **kwargs):
        # 'placeholder' option presence is required for 'allowClear'
        params = {"placeholder": ""}
        if self.unescape_html:
            params["makeHtml"] = True
        if not field.flags.required:
            params["allowClear"] = True

        css_class = kwargs.setdefault("class_", "")
        if "js-widget" not in css_class:
            css_class += " js-widget"
            kwargs["class_"] = css_class

        kwargs.setdefault("data-init-with", self.js_init)
        kwargs["data-init-params"] = json.dumps(params)
        return Select.__call__(self, field, **kwargs)

    def render_view(self, field, **kwargs):
        labels = [str(label) for v, label, checked in field.iter_choices() if checked]
        return "; ".join(labels)

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        if value is None:
            return Markup("<option></option>")
        return Select.render_option(value, label, selected, **kwargs)


class Select2Ajax:
    """Ad-hoc select widget based on Select2.

    The code below is probably very fragile, since it depends on the internal
    structure of a Select2 widget.

    :param format_result: `formatResult` arg of Select2. Must be a valid
    javascript reference (like `Abilian.my_format_function`)

    :param format_selection: like `format_result`, but for Select2's
    `formatSelection`

    :param values_builder: an optional function, which given a data iterable of
    values, returns a list of dict items suitable for widget initial data;
    generally needed when using custom format_result/format_selection.
    """

    def __init__(
        self,
        template="widgets/select2ajax.html",
        multiple=False,
        format_result=None,
        format_selection=None,
        values_builder=None,
    ):
        self.template = template
        self.multiple = multiple
        self.values_builder = (
            values_builder
            if callable(values_builder)
            else lambda data: [{"id": o.id, "text": o.name} for o in data if o]
        )
        self.s2_params = {"multiple": self.multiple}

        if format_result:
            self.s2_params["formatResult"] = format_result
        if format_selection:
            self.s2_params["formatSelection"] = format_selection

    def process_formdata(self, valuelist):
        """field helper: as of select2 3.x, multiple values are passed as a
        single string."""
        if self.multiple:
            valuelist = valuelist[0].split(",")
        return valuelist

    def __call__(self, field, **kwargs):
        """Render widget."""
        if self.multiple:
            kwargs["multiple"] = True

        css_class = kwargs.setdefault("class_", "")
        if "class" in kwargs:
            css_class += kwargs.pop("class", "")
            kwargs["class_"] = css_class

        if "js-widget" not in css_class:
            css_class += " js-widget"
            kwargs["class_"] = css_class

        s2_params = {}
        s2_params.update(self.s2_params)

        if field.ajax_source:
            s2_params["ajax"] = {"url": str(field.ajax_source)}

        s2_params["placeholder"] = str(field.label.text)

        if not field.flags.required:
            s2_params["allowClear"] = True

        data = field.data
        if not self.multiple:
            data = [data]

        values = self.values_builder(data)
        input_value = ",".join(str(o.id) for o in data if o)
        data_node_id = None

        json_data = {}
        if values:
            json_data["values"] = values
            data_node_id = s2_params["dataNodeId"] = f"{field.id}-json"

        kwargs["data-init-with"] = "select2ajax"
        kwargs["data-init-params"] = json.dumps(s2_params)

        extra_args = Markup(html_params(**kwargs))

        ctx = {
            "field": field,
            "name": field.name,
            "id": field.id,
            "input_value": input_value,
            "json_data": json_data,
            "required": not field.allow_blank,
            "data_node_id": data_node_id,
            "extra_args": extra_args,
        }
        return Markup(render_template(self.template, **ctx))



"""Configuration and injectable fixtures for Pytest.

Can be reused (and overriden) by adding::

   pytest_plugins = ['abilian.testing.fixtures']

to your `conftest.py`.

Replaced the too-complex UnitTest-based testing framework.
-> DI and functions over complex inheritance hierarchies FTW!
"""
from __future__ import annotations

import typing
from typing import Any, Iterator

from flask import Flask
from flask.ctx import AppContext, RequestContext
from flask.testing import FlaskClient
from pytest import fixture
from sqlalchemy.orm import Session

if typing.TYPE_CHECKING:
    from abilian.core.models.subjects import User
    from abilian.core.sqlalchemy import SQLAlchemy


class TestConfig:
    TESTING = True
    DEBUG = True
    SECRET_KEY = "SECRET"
    SERVER_NAME = "localhost.localdomain"
    CELERY_ALWAYS_EAGER = True  # run tasks locally, no async
    CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
    MAIL_SENDER = "tester@example.com"
    SITE_NAME = "Abilian Test"
    # WTF_CSRF_ENABLED = True
    WTF_CSRF_ENABLED = False
    BABEL_ACCEPT_LANGUAGES = ["en", "fr"]


@fixture
def config() -> type:
    return TestConfig


@fixture
def app(config: Any) -> Flask:
    # We currently return a fresh app for each test.
    # Using session-scoped app doesn't currently work.
    # Note: the impact on speed is minimal.
    from abilian.app import create_app

    return create_app(config=config)


@fixture
def app_context(app: Flask) -> Iterator[AppContext]:
    with app.app_context() as ctx:
        yield ctx


@fixture
def test_request_context(app: Flask) -> Iterator[RequestContext]:
    with app.test_request_context() as ctx:
        yield ctx


# @fixture
# def req_ctx(app: Flask) -> Iterator[RequestContext]:
#     with app.test_request_context() as _req_ctx:
#         yield _req_ctx


@fixture
def db(app_context: AppContext) -> Iterator[SQLAlchemy]:
    """Return a fresh db for each test."""
    from abilian.core.extensions import db
    from abilian.testing.util import (
        cleanup_db,
        ensure_services_started,
        stop_all_services,
    )

    stop_all_services(app_context.app)
    ensure_services_started(["repository", "session_repository"])

    cleanup_db(db)
    db.create_all()
    yield db

    db.session.remove()
    cleanup_db(db)
    stop_all_services(app_context.app)


@fixture
def session(db: SQLAlchemy) -> Session:
    return db.session


@fixture
def db_session(db: SQLAlchemy) -> Session:
    return db.session


# @fixture
# def client(app: Flask) -> FlaskClient:
#     """Return a Web client, used for testing."""
#     return app.test_client()


@fixture
def user(db: SQLAlchemy) -> User:
    from abilian.core.models.subjects import User

    user = User(
        first_name="Joe",
        last_name="Test",
        email="test@example.com",
        password="test",
        can_login=True,
    )
    db.session.add(user)
    db.session.flush()
    return user


@fixture
def admin_user(db: SQLAlchemy) -> User:
    from abilian.core.models.subjects import User

    user = User(
        first_name="Jim",
        last_name="Admin",
        email="admin@example.com",
        password="admin",
        can_login=True,
    )
    user.is_admin = True
    db.session.add(user)
    db.session.flush()
    return user


@fixture
def login_user(user: User, client: FlaskClient) -> User:
    with client.session_transaction() as session:
        session["_user_id"] = user.id

    return user


@fixture
def login_admin(admin_user: User, client: FlaskClient) -> User:
    with client.session_transaction() as session:
        session["_user_id"] = admin_user.id

    return admin_user



import abc
from functools import total_ordering

import sqlalchemy as sa
import sqlalchemy.orm

from abilian.core.entities import Entity
from abilian.core.extensions import db

from .base import IdMixin, Model

#: backref attribute on tagged elements
TAGS_ATTR = "__tags__"


class SupportTagging(metaclass=abc.ABCMeta):
    pass


def register(cls):
    """Register an :class:`~.Entity` as a taggable class.

    Can be used as a class decorator:

    .. code-block:: python

        @tag.register
        class MyContent(Entity):
            ....
    """
    if not issubclass(cls, Entity):
        raise ValueError("Class must be a subclass of abilian.core.entities.Entity")

    SupportTagging.register(cls)
    return cls


def supports_tagging(obj):
    """
    :param obj: a class or instance
    """
    if isinstance(obj, type):
        return issubclass(obj, SupportTagging)

    if not isinstance(obj, SupportTagging):
        return False

    if obj.id is None:
        return False

    return True


entity_tag_tbl = db.Table(
    "entity_tags",
    Model.metadata,
    sa.Column("tag_id", sa.Integer, sa.ForeignKey("tag.id", ondelete="CASCADE")),
    sa.Column("entity_id", sa.Integer, sa.ForeignKey(Entity.id, ondelete="CASCADE")),
    sa.UniqueConstraint("tag_id", "entity_id"),
)


@total_ordering
class Tag(IdMixin, Model):
    """Tags are text labels that can be attached to :class:`entities.Entity`.

    They are namespaced, so that independent group of tags can be
    defined in the application. The default namespace is `"default"`.
    """

    __tablename__ = "tag"

    #: namespace
    ns = sa.Column(
        sa.UnicodeText(), nullable=False, default="default", server_default="default"
    )

    #: Label visible to the user
    label = sa.Column(sa.UnicodeText(), nullable=False)

    #: :class:`entities <.Entity>` attached to this tag
    entities = sa.orm.relationship(
        Entity,
        collection_class=set,
        secondary=entity_tag_tbl,
        backref=sa.orm.backref(TAGS_ATTR, collection_class=set),
    )

    __mapper_args__ = {"order_by": label}

    __table_args__ = (
        sa.UniqueConstraint(ns, label),
        # namespace is not empty and is not surrounded by space characters
        sa.CheckConstraint(sa.sql.and_(sa.sql.func.trim(ns) == ns, ns != "")),
        # label is not empty and is not surrounded by space characters
        sa.CheckConstraint(sa.sql.and_(sa.sql.func.trim(label) == label, label != "")),
    )

    def __str__(self):
        return self.label

    def __lt__(self, other):
        return str(self).lower().__lt__(str(other).lower())

    def __repr__(self):
        cls = self.__class__
        return (
            "<{mod}.{cls} id={t.id!r} ns={t.ns!r} "
            "label={t.label!r} at 0x{addr:x}>".format(
                mod=cls.__module__, cls=cls.__name__, t=self, addr=id(self)
            )
        )



import logging
from typing import TYPE_CHECKING, Any, Optional, Tuple, Type

from flask import Request
from sqlalchemy import func, orm
from sqlalchemy.sql.expression import or_

from abilian.core.entities import EntityQuery
from abilian.core.models import Model

if TYPE_CHECKING:
    from abilian.web.frontend import Module

logger = logging.getLogger(__name__)


class UNDEFINED:
    pass


_UNDEFINED = UNDEFINED()


class BaseCriterion:
    form_default_value: Any = _UNDEFINED
    """
    Values to set by default when adding a filter. The provided value(s) must be
    the ones used in html, not in python.

    Subclasses can also define property.
    """

    def __init__(
        self, name: str, label: str = "", form_default_value: Any = _UNDEFINED
    ) -> None:
        self.name = name
        self.label = label

        if form_default_value is not _UNDEFINED:
            self.form_default_value = form_default_value  # optionnal default value

    # model must be set before calling any method or property
    _model = None

    def _get_model(self):
        if self._model is None:
            raise ValueError("Model not set")
        return self._model

    def _set_model(self, model: Type[Model]) -> None:
        if self._model is not None:
            raise ValueError("Model already set")

        self._model = model

    model = property(_get_model, _set_model)

    def filter(self, query, module, request, searched_text, *args, **kwargs):
        raise NotImplementedError

    @property
    def has_form_filter(self):
        return True

    @property
    def form_filter_type(self):
        raise NotImplementedError

    @property
    def form_filter_args(self):
        raise NotImplementedError

    @property
    def form_unset_value(self):
        raise NotImplementedError

    @property
    def has_form_default_value(self):
        return self.form_default_value is not _UNDEFINED

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class TextSearchCriterion(BaseCriterion):
    """Fulltext search on given attributes."""

    def __init__(
        self,
        name: str,
        label: str = "",
        attributes: Optional[Tuple[str, str]] = None,
        search_fmt: str = "%{q}%",
    ) -> None:
        super().__init__(name, label)
        self.attributes = dict.fromkeys(
            attributes if attributes is not None else (name,)
        )
        self._attributes_prepared = False

        if isinstance(search_fmt, str):
            search_fmt = [search_fmt]

        self.search_fmt = search_fmt

    def _prepare_attributes(self):
        to_del = []

        for attr_name in self.attributes:
            name = attr_name
            val = self.attributes[name] = {}

            if "." in attr_name:
                # related model
                rel_attr_name, name = attr_name.split(".", 1)
                model, attr = self.get_rel_attr(attr_name, self.model)
            else:
                rel_attr_name = None
                model = None
                attr = getattr(self.model, attr_name, None)

            if attr is None:
                logger.debug(
                    'Model: "%s", could not find "%s"',
                    self.model.__class__.__name__,
                    attr_name,
                )
                to_del.append(attr_name)
            else:
                val.update(
                    {
                        "attr": attr,
                        "name": name,
                        "model": model,
                        "rel_attr_name": rel_attr_name,
                    }
                )

        for k in to_del:
            del self.attributes[k]

        self._attributes_prepared = True

    def filter(
        self,
        query: EntityQuery,
        module: "Module",
        request: Request,
        searched_text: str,
        *args,
        **kwargs,
    ) -> EntityQuery:

        if not searched_text:
            return query

        if not self._attributes_prepared:
            self._prepare_attributes()

        clauses = []
        has_joins = False

        for attr_name, val in self.attributes.items():
            if self.is_excluded(attr_name, request):
                continue

            attr = val["attr"]

            if val["model"] is not None:
                # related model - generate an alias, required when searched model has
                # more than one relationship with another model
                model = orm.aliased(val["model"])
                attr = getattr(model, attr.key)

                join_attr = getattr(module.managed_class, val["rel_attr_name"])
                query = query.outerjoin(model, join_attr)

                has_joins = True

            # TODO: gérer les accents
            for fmt in self.search_fmt:
                like_txt = fmt.format(q=searched_text)
                clauses.append(func.lower(attr).like(like_txt))

        if clauses:
            query = query.filter(or_(*clauses)).distinct()

        if has_joins:
            query = query.reset_joinpoint()

        return query

    def get_rel_attr(self, attr_name, model):
        """For a related attribute specification, returns (related model,
        attribute).

        Returns (None, None) if model is not found, or (model, None) if
        attribute is not found.
        """
        rel_attr_name, attr_name = attr_name.split(".", 1)
        rel_attr = getattr(self.model, rel_attr_name, None)
        rel_model = None
        attr = None

        if rel_attr is not None:
            rel_model = rel_attr.property.mapper.class_
            attr = getattr(rel_model, attr_name, None)

        return rel_model, attr

    def is_excluded(self, attr_name: str, request: Request):
        """To be overriden by subclasses that want to filter searched
        attributes."""
        return False

    @property
    def has_form_filter(self) -> bool:
        return False


class TextCriterion(TextSearchCriterion):
    def __init__(self, name, label="", attributes=None, search_fmt="%{q}%"):
        super().__init__(name, label, attributes, search_fmt)

    def filter(self, query, module, request, searched_text, *args, **kwargs):
        my_searched_text = request.values.get(self.name, "").strip()
        if my_searched_text:
            return super().filter(
                query, module, request, my_searched_text.lower(), *args, **kwargs
            )
        else:
            return query

    @property
    def has_form_filter(self):
        return True

    @property
    def form_filter_type(self):
        return "text"

    @property
    def form_filter_args(self):
        return []

    @property
    def form_unset_value(self):
        return ""



""""""
import sqlalchemy as sa
import sqlalchemy.exc
from pytest import mark, raises

from abilian.testing.util import path_from_url
from abilian.web import url_for

from .models import BaseVocabulary, Vocabulary
from .service import vocabularies

# DefaultVoc = Vocabulary('defaultstates', group='', label='States')

# XXX: we need to use globals for now.
# Classes are global state in Python.
Voc = Vocabulary("voc", label="Voc", group="")
PriorityVoc = Vocabulary("priorities", label="Priorities", group="")
StateVoc = Vocabulary("defaultstates", group="", label="States")
DocCatVoc = Vocabulary("categories", group="documents", label="Categories")


def test_vocabulary_factory(session):
    assert Voc.__name__ == "VocabularyVoc"
    assert Voc.__tablename__ == "vocabulary_voc"
    assert Voc.Meta.name == "voc"
    assert Voc.Meta.label == "Voc"
    assert Voc.Meta.group == ""
    assert issubclass(Voc, BaseVocabulary)


def test_vocabularies(session):
    # test registered vocabularies
    assert vocabularies.vocabularies == {Voc, PriorityVoc, StateVoc, DocCatVoc}

    grouped = vocabularies.grouped_vocabularies
    assert set(grouped[""]) == {Voc, StateVoc, PriorityVoc}
    assert grouped["documents"] == [DocCatVoc]

    assert vocabularies.get_vocabulary("priorities") is PriorityVoc
    assert vocabularies.get_vocabulary("priorities", "nogroup") is None
    assert vocabularies.get_vocabulary("categories", "documents") is DocCatVoc


def test_items(db, session):
    db.create_all()

    IMMEDIATE = PriorityVoc(label="Immediate", position=0)
    NORMAL = PriorityVoc(label="Normal", position=3, default=True)
    URGENT = PriorityVoc(label="Urgent", position=1)
    HIGH = PriorityVoc(label="High", position=2)
    items = (IMMEDIATE, NORMAL, URGENT, HIGH)
    for item in items:
        session.add(item)
    session.flush()

    # test position=4 set automatically; Label stripped
    low_item = PriorityVoc(label=" Low  ")
    session.add(low_item)
    session.flush()
    assert low_item.position == 4
    assert low_item.label == "Low"

    # test strip label on update
    IMMEDIATE.label = "  Immediate  "
    session.flush()
    assert IMMEDIATE.label == "Immediate"

    # test default ordering
    default_ordering = ["Immediate", "Urgent", "High", "Normal", "Low"]
    query = PriorityVoc.query
    assert [i.label for i in query.active().all()] == default_ordering

    # no default ordering when using .values(): explicit ordering required
    query = PriorityVoc.query.active().order_by(PriorityVoc.position.asc())
    assert [i.label for i in query.values(PriorityVoc.label)] == default_ordering

    # test db-side constraint for non-empty labels
    with raises(sa.exc.IntegrityError):
        with session.begin_nested():
            v = PriorityVoc(label="   ", position=6)
            session.add(v)
            session.flush()

    # test unique labels constraint
    with raises(sa.exc.IntegrityError):
        with session.begin_nested():
            v = PriorityVoc(label="Immediate")
            session.add(v)
            session.flush()

    # test unique position constraint
    with raises(sa.exc.IntegrityError):
        with session.begin_nested():
            v = PriorityVoc(label="New one", position=1)
            session.add(v)
            session.flush()

    # test by_position without results
    item = PriorityVoc.query.by_position(42)
    assert item is None

    # test by_position() and active()
    item = PriorityVoc.query.by_position(URGENT.position)
    assert item is URGENT

    item.active = False
    expected = ["Immediate", "High", "Normal", "Low"]
    assert [i.label for i in PriorityVoc.query.active().all()] == expected
    assert PriorityVoc.query.active().by_position(URGENT.position) is None

    # test by_label()
    item = PriorityVoc.query.by_label(NORMAL.label)
    assert item is NORMAL


@mark.skip
def test_admin_panel_reorder(app, db, session, client, test_request_context):
    db.create_all()
    items = [
        Voc(label="First", position=0),
        Voc(label="Second", position=2),
        Voc(label="Third", position=3),
    ]

    for i in items:
        session.add(i)
    session.commit()

    first, second, third = items
    url = url_for("admin.vocabularies")
    base_data = {"Model": Voc.Meta.name}
    data = {"down": first.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert path_from_url(r.location) == "/admin/vocabularies"
    assert Voc.query.order_by(Voc.position).all() == [second, first, third]

    data = {"up": first.id, "return_to": "group"}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert path_from_url(r.location) == "/admin/vocabularies/_/"
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {"up": first.id, "return_to": "model"}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert path_from_url(r.location) == "/admin/vocabularies/_/defaultstates/"
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {"down": third.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert path_from_url(r.location) == "/admin/vocabularies"
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {"up": third.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert path_from_url(r.location) == "/admin/vocabularies"
    assert Voc.query.order_by(Voc.position).all() == [first, third, second]
